from qiskit.quantum_info import SparsePauliOp

from qiskit_nature.second_q.formats.molecule_info import MoleculeInfo
from qiskit_nature.second_q.drivers import PySCFDriver
from qiskit_nature.second_q.mappers import ParityMapper, QubitMapper
from qiskit_nature.second_q.problems import ElectronicStructureProblem

from qlauncher.base import Problem


class Molecule(Problem):
    def __init__(self, instance: MoleculeInfo, mapper: QubitMapper | None = None, basis_set: str = 'STO-6G', instance_name: str = 'unnamed') -> None:
        self.basis_set = basis_set
        self.mapper = ParityMapper() if mapper is None else mapper
        self.problem: ElectronicStructureProblem = self._get_problem(instance)
        self.operator: SparsePauliOp = self._get_operator(self.problem)
        super().__init__(instance, instance_name)

    @staticmethod
    def from_preset(instance_name: str) -> "Molecule":
        match instance_name:
            case 'H2':
                instance = MoleculeInfo(["H", "H"], [(0.0, 0.0, 0.0), (0.0, 0.0, 0.74)])
            case _:
                raise ValueError(f"Molecule {instance_name} not supported, currently you can use: 'H2'")
        return Molecule(instance, instance_name=instance_name)

    def _get_problem(self, molecule) -> ElectronicStructureProblem:
        driver = PySCFDriver.from_molecule(molecule, basis=self.basis_set)
        problem = driver.run()
        return problem

    def _get_operator(self, problem) -> SparsePauliOp:
        self.mapper.num_particles = problem.num_particles
        hamiltonian = self.mapper.map(problem.hamiltonian.second_q_op())
        if not isinstance(hamiltonian, SparsePauliOp):
            raise ValueError
        return hamiltonian



"""This module contains the Graph Coloring class."""

import pickle
import networkx as nx
from random import randint
import matplotlib.pyplot as plt
from qlauncher.base import Problem


class GraphColoring(Problem):
    """
    Class for Graph Coloring Problem which is a combinatorial problem involving assigning labels to vertices of the graph such that no two adjacent
    vertices share the same label.

    Attributes:
        instance (nx.Graph): The graph for which the coloring problem is to be solved.


    """

    def __init__(
        self,
        instance: nx.Graph,
        num_colors: int,
        instance_name: str = "unnamed",
    ) -> None:
        super().__init__(instance=instance, instance_name=instance_name)
        self.pos = None
        self.num_colors = num_colors

    @property
    def setup(self) -> dict:
        return {"instance_name": self.instance_name}

    @staticmethod
    def from_preset(instance_name: str) -> "GraphColoring":
        match instance_name:
            case "default":
                graph = nx.petersen_graph()
                gc = GraphColoring(graph, instance_name=instance_name, num_colors=3)
                gc.pos = nx.shell_layout(gc.instance, nlist=[list(range(5, 10)), list(range(5))])
                return gc
            case "small":
                graph = nx.Graph()
                graph.add_edges_from([(0, 1), (1, 2), (0, 2), (3, 2)])
                return GraphColoring(graph, instance_name=instance_name, num_colors=3)

    @staticmethod
    def from_file(path: str) -> "GraphColoring":
        with open(path, 'rb') as f:
            graph, num_colors = pickle.load(f)
        return GraphColoring(graph, instance_name=path, num_colors=num_colors)

    def to_file(self, path: str):
        with open(path, 'wb') as f:
            pickle.dump((self.instance, self.num_colors), f, pickle.HIGHEST_PROTOCOL)

    def _get_path(self) -> str:
        return f"{self.name}@{self.instance_name}"

    def visualize(self, solution: list[int] | None = None):
        if self.pos is None:
            self.pos = nx.spring_layout(self.instance, seed=42)  # set seed for same node graphs in plt
        plt.figure(figsize=(8, 6))
        if solution is not None:
            nx.draw_networkx_nodes(self.instance, self.pos, node_size=500, node_color=solution, cmap="Accent")
            colors = []
            for n1,n2 in self.instance.edges:
                if solution[n1] == solution[n2]:
                    colors.append("r")
                else:
                    colors.append("gray")
            nx.draw_networkx_edges(self.instance, self.pos, edge_color=colors,width=4)
        else:
            nx.draw_networkx_nodes(self.instance, self.pos, node_size=500, node_color="skyblue")
        nx.draw_networkx_edges(self.instance, self.pos, edge_color="gray")
        nx.draw_networkx_labels(self.instance, self.pos, font_size=10, font_weight="bold")
        plt.title(f"Graph {self.num_colors}-Coloring Problem Instance Visualization")
        plt.show()

    @staticmethod
    def generate_graph_coloring_instance(num_vertices: int, edge_probability: int, num_colors: int) -> "GraphColoring":
        graph = nx.gnp_random_graph(num_vertices, edge_probability)
        return GraphColoring(graph, num_colors=num_colors)

    @staticmethod
    def randomly_choose_a_graph(num_colors: int) -> "GraphColoring":
        graphs = nx.graph_atlas_g()
        graph = graphs[randint(0, len(graphs) - 1)]
        return GraphColoring(graph, num_colors=num_colors)



import concurrent.futures
from typing import Any, Literal
from collections.abc import Callable
from qlauncher.base import Algorithm


class Task:
    def __init__(self, func: Callable, args: tuple[Any] | None = None, kwargs: dict[str, Any] | None = None, num_output: int = 1):
        if args is None:
            args = tuple()
        if kwargs is None:
            kwargs = {}
        self.func = func
        self.dependencies: list[Task] = [arg for arg in args if isinstance(arg, Task)]
        self.dependencies.extend([value for value in kwargs.values() if isinstance(value, Task)])
        self.args = args
        self.kwargs = kwargs
        self.done = False
        self.result = None
        self.num_output = num_output
        self.subtasks: list[SubTask] = []

    def run(self):
        binded_args = [arg.result if isinstance(arg, Task) else arg for arg in self.args]
        binded_kwargs = {key: (value.result if isinstance(value, Task) else value) for key, value in self.kwargs.items()}
        self.result = self.func(*binded_args, **binded_kwargs)
        self.done = True

    def is_ready(self):
        return all(map(lambda x: x.done, self.dependencies))

    def __iter__(self):
        for i in range(self.num_output):
            yield SubTask(self, i)


class SubTask(Task):
    def __init__(self, task: Task, index: int):
        self.task = task
        self.index = index

    @property
    def result(self):
        return self.task.result[self.index]

    @property
    def done(self):
        return self.task.done


class Workflow(Algorithm):
    _algorithm_format = 'none'

    def __init__(self, tasks: list[Task], input_task: Task, output_task: Task, input_format: str = 'none'):
        self.tasks = tasks
        self.input_task = input_task
        self.output_task = output_task
        self._algorithm_format = input_format

    def run(self, problem, backend, formatter):
        input_result = formatter(problem)
        self.input_task.result = input_result
        with concurrent.futures.ThreadPoolExecutor() as executor:
            _execute_workflow(self.tasks, executor)
        return self.output_task.result


class WorkflowManager:
    def __init__(self, manager: Literal['ql', 'prefect', 'airflow'] = 'ql'):
        self.tasks: list[Task] = []
        self.manager = manager
        self.input_task: Task | None = None
        self.input_task_format: str = 'none'
        self.output_task: Task | None = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def task(self, func, args: tuple | None = None, kwargs: dict | None = None, num_output=None) -> Task:
        args = args or tuple()
        kwargs = kwargs or dict()
        new_task = Task(func, args, kwargs, num_output=num_output)
        self.tasks.append(new_task)
        return new_task

    def __call__(self, input_value=None, /):
        if self.input_task:
            self.input_task.result = input_value
        with concurrent.futures.ThreadPoolExecutor() as executor:
            _execute_workflow(self.tasks, executor)
        if self.output_task:
            return self.output_task.result

    def print_dag(self):
        for task in self.tasks:
            dep_names = [dep.func.__name__ for dep in task.dependencies]
            print(f"{task.func.__name__} -> {dep_names}")

    def input(self, format: str = 'none'):
        self.input_task = Task(func=None)
        self.input_task.done = True
        self.input_task_format = format
        return self.input_task

    def output(self, task: Task):
        self.output_task = task

    def to_workflow(self) -> Workflow:
        return Workflow(self.tasks, self.input_task, self.output_task, input_format=self.input_task_format)


def _execute_workflow(tasks: list[Task], executor: concurrent.futures.Executor, max_iterations: int | None = None):
    remaining_tasks = set(tasks)
    max_iterations: int = max_iterations or len(remaining_tasks)
    iteration = 0
    for _ in range(max_iterations):
        ready_tasks = list(filter(Task.is_ready, remaining_tasks))

        if len(ready_tasks) < 1:
            if remaining_tasks:
                raise RuntimeError("Cycle or error in tasks.")
            return

        futures = {executor.submit(task.run): task for task in ready_tasks}
        for future in concurrent.futures.as_completed(futures):
            if future.exception():
                raise future.exception()

        for t in ready_tasks:
            remaining_tasks.remove(t)

        if iteration > max_iterations:
            raise RuntimeError("Processing take too much iterations")
        iteration += 1



"""  This module contains the QATM class."""
import os

from typing import Any

import numpy as np
import pandas as pd

from qlauncher.base import Problem


class QATM(Problem):
    """ 
    Class for Quantum Aircraft Traffic Management (QATM) problem.

    This class represents Quantum Aircraft Traffic Management (QATM) problem which is a combinatorial optimization problem
    that involves scheduling a set of aircrafts on a set of manouvers. Each aircraft consists of a sequence of manouvers
    that must be performed. The objective is to find a schedule that minimizes the number of collisions between aircrafts.
    The class contains an instance of the problem, so it can be passed into QLauncher.

    Attributes:
        onehot (str): The one-hot encoding used for the problem.
        instance (dict): The instance of the problem.
        instance_name (str | None): The name of the instance.
        instance_path (str): The path to the instance file.

    """

    def __init__(self, onehot: str = 'exact', instance: Any = None, instance_name: str | None = None,
                 optimization_problem: bool = False) -> None:
        super().__init__(instance=instance, instance_name=instance_name if instance_name is not None else "QATM")
        self.onehot = onehot
        self.optimization_problem = optimization_problem

    @property
    def setup(self) -> dict:
        return {
            'onehot': self.onehot,
            'instance_name': self.instance_name
        }

    def _get_path(self) -> str:
        return f'{self.name}@{self.instance_name.split(".", 1)[0]}'

    @classmethod
    def from_file(cls, path: str, instance_name: str = "QATM"):
        cm_path = os.path.join(path, 'CM_' + instance_name)
        aircrafts_path = os.path.join(
            path, 'aircrafts_' + instance_name)

        instance = {'cm': np.loadtxt(cm_path), 'aircrafts': pd.read_csv(aircrafts_path, delimiter=' ', names=['manouver', 'aircraft'])}
        return QATM('exact', instance, instance_name)

    def analyze_result(self, result: dict):
        """
        Analyzes the result in terms of collisions and violations of onehot constraint.

        Parameters:
            result (dict): A dictionary where keys are bitstrings and values are probabilities.

        Returns:
            dict: A dictionary containing collisions, onehot violations, and changes as ndarrays.
        """
        keys = list(result.keys())
        vectorized_result = (np.fromstring("".join(keys), 'u1') - ord('0')).reshape(len(result), -1)

        cm = self.instance['cm'].copy().astype(int)
        np.fill_diagonal(cm, 0)
        collisions = np.einsum('ij,ij->i', vectorized_result @ cm, vectorized_result) / 2

        df = pd.DataFrame(vectorized_result.transpose())
        df['aircraft'] = self.instance['aircrafts']['aircraft']
        onehot_violations = (df.groupby(by='aircraft').sum() != 1).sum(axis=0).to_numpy()

        df['manouver'] = self.instance['aircrafts']['manouver']
        no_changes = df[df['aircraft'] == df['manouver']]
        changes = (len(no_changes) - no_changes.drop(['manouver', 'aircraft'], axis=1).sum()).to_numpy().astype(int)
        changes[onehot_violations != 0] = -1

        at_least_one = (df.loc[:, df.columns != 'manouver'].groupby('aircraft').sum() > 0).all().to_numpy().astype(int)

        return {
            'collisions': collisions,
            'onehot_violations': onehot_violations,
            'changes': changes,
            'at_least_one': at_least_one
        }



""" Algorithms for Qiskit routines """
from datetime import datetime
from collections.abc import Callable, Iterable
from typing import Any, Literal
import statistics
import numpy as np
from scipy.optimize import minimize

import qiskit_algorithms
from qiskit_algorithms import optimizers
from qiskit_algorithms.minimum_eigensolvers.diagonal_estimator import _evaluate_sparsepauli as evaluate_energy

from qiskit import QuantumCircuit
from qiskit.circuit.library import PauliEvolutionGate, QAOAAnsatz, efficient_su2

from qiskit.primitives import PrimitiveResult, SamplerPubResult, BaseSamplerV1, BaseEstimatorV1
from qiskit.primitives.containers import BitArray
from qiskit.primitives.base.base_primitive import BasePrimitive

from qiskit.quantum_info import SparsePauliOp

from qiskit_nature.second_q.algorithms.ground_state_solvers import GroundStateEigensolver
from qiskit_nature.second_q.problems import EigenstateResult

from qlauncher.base import Problem, Algorithm, Result
from qlauncher.base.base import Backend
from qlauncher.routines.cirq import CirqBackend
from qlauncher.routines.qiskit.backends.qiskit_backend import QiskitBackend
from qlauncher.problems import Molecule


class QiskitOptimizationAlgorithm(Algorithm):
    """ Abstract class for Qiskit optimization algorithms """

    def make_tag(self, problem: Problem, backend: QiskitBackend) -> str:
        tag = problem.__class__.__name__ + '-' + \
            backend.__class__.__name__ + '-' + \
            self.__class__.__name__ + '-' + \
            datetime.today().strftime('%Y-%m-%d')
        return tag

    def get_processing_times(self, tag: str, primitive: BasePrimitive) -> None | tuple[list, list, int]:
        timestamps = []
        usages = []
        qpu_time = 0
        if hasattr(primitive, 'session'):
            jobs = primitive.session.service.jobs(limit=None, job_tags=[tag])
            for job in jobs:
                m = job.metrics()
                timestamps.append(m['timestamps'])
                usages.append(m['usage'])
                qpu_time += m['usage']['quantum_seconds']
        return timestamps, usages, qpu_time


def commutator(op_a: SparsePauliOp, op_b: SparsePauliOp) -> SparsePauliOp:
    """ Commutator """
    return op_a @ op_b - op_b @ op_a


def int_to_bitstring(number: int, total_bits: int):
    return np.binary_repr(number, total_bits)[::-1]


def cvar_cost(probs_values: Iterable[tuple[float, float]], alpha: float):
    """CVar cost function to be used instead of mean for qaoa training"""
    if not (alpha > 0 and alpha <= 1):
        raise ValueError("Alpha must be in range (0,1]")
    cvar, acc = 0.0, 0.0
    for p, v in sorted(probs_values, key=lambda x: x[1]):
        if acc >= alpha:
            break
        acc += p
        cvar += v * min(p, alpha-acc)
    return cvar / alpha


class QAOA(QiskitOptimizationAlgorithm):
    """Algorithm class with QAOA.

    Args:
        p (int): The number of QAOA steps. Defaults to 1.
        optimizer (Optimizer | None): Optimizer used during algorithm runtime. If set to `None` turns into COBYLA. Defaults to None,
        alternating_ansatz (bool): Whether to use an alternating ansatz. Defaults to False. If True, it's recommended to provide a mixer_h to alg_kwargs.
        aux: Auxiliary input for the QAOA algorithm.
        **alg_kwargs: Additional keyword arguments for the base class.

    Attributes:
        name (str): The name of the algorithm.
        aux: Auxiliary input for the QAOA algorithm.
        p (int): The number of QAOA steps.
        optimizer (Optimizer): Optimizer used during algorithm runtime.
        alternating_ansatz (bool): Whether to use an alternating ansatz.
        parameters (list): List of parameters for the algorithm.
        mixer_h (SparsePauliOp | None): The mixer Hamiltonian.

    """
    _algorithm_format = 'hamiltonian'

    def __init__(
            self,
            p: int = 1,
            optimization_method: Literal['COBYLA'] = "COBYLA",
            max_evaluations: int = 100,
            training_aggregation_method: Literal['mean', 'cvar'] = 'mean',
            cvar_alpha: float = 1,
            alternating_ansatz: bool = False,
            aux=None,
            **alg_kwargs):
        super().__init__(**alg_kwargs)
        self.name: str = 'qaoa'
        self.aux = aux
        self.p: int = p

        self.optimization_method = optimization_method
        self.max_evaluations = max_evaluations
        self.training_aggregation_method = training_aggregation_method
        self.cvar_alpha = cvar_alpha

        self.alternating_ansatz: bool = alternating_ansatz
        self.parameters = ['p']
        self.mixer_h: SparsePauliOp | None = None
        self.initial_state: QuantumCircuit | None = None

    @property
    def setup(self) -> dict:
        return {
            'aux': self.aux,
            'p': self.p,
            'parameters': self.parameters,
            'arg_kwargs': self.alg_kwargs
        }

    def _get_optimized_circuit_params(self, circuit: QuantumCircuit, hamiltonian: SparsePauliOp, backend: QiskitBackend | CirqBackend) -> tuple[np.ndarray, list[float]]:
        """
        Optimize circuit params

        Args:
            circuit (QuantumCircuit): QAOA circuit to be optimized
            backend (Backend): Backend containing sampler

        Returns:
            tuple[QuantumCircuit,list[float]]: Circuit with optimal param values applied, energy history
        """
        costs = []

        def cost_fn(params: np.ndarray):
            job = backend.sampler.run([(circuit, params)])
            results = job.result()[0].data.meas.get_int_counts()
            shots = sum(results.values())

            probs_with_costs = {state: (count/shots, np.real(evaluate_energy(state, hamiltonian)))
                                for state, count in results.items()}

            cost = (sum(prob*cost for prob, cost in probs_with_costs.values())
                    if self.training_aggregation_method == 'mean' else
                    cvar_cost(probs_with_costs.values(), self.cvar_alpha))
            costs.append(cost)
            return cost

        res = minimize(
            cost_fn,
            np.array([np.pi]*(len(circuit.parameters)//2) + [np.pi / 2]*(len(circuit.parameters)//2)),
            method=self.optimization_method,
            tol=1e-2,
            options={
                'maxiter': self.max_evaluations
            }
        )

        return res.x, costs

    def run(self, problem: Problem, backend: Backend, formatter: Callable) -> Result:
        """ Runs the QAOA algorithm """

        if not (isinstance(backend, QiskitBackend) or isinstance(backend, CirqBackend)):
            raise ValueError('Backend should be CirqBackend, QiskitBackend or subclass.')

        hamiltonian: SparsePauliOp = formatter(problem)

        if self.alternating_ansatz:
            if self.mixer_h is None:
                self.mixer_h = formatter.get_mixer_hamiltonian(problem)
            if self.initial_state is None:
                self.initial_state = formatter.get_QAOAAnsatz_initial_state(
                    problem)

        # Cirq translation issues if we use QAOAAnsatz() by itself without appending it to a QuantumCircuit
        circuit = QuantumCircuit(hamiltonian.num_qubits)
        circuit.append(QAOAAnsatz(cost_operator=hamiltonian, reps=self.p).to_instruction(), range(hamiltonian.num_qubits))

        circuit.measure_all()

        opt_params, costs = self._get_optimized_circuit_params(circuit, hamiltonian, backend)

        job = backend.sampler.run([(circuit, opt_params)])
        results = job.result()[0].data.meas.get_int_counts()

        final_energies = {
            int_to_bitstring(k, circuit.num_qubits): np.real(evaluate_energy(k, hamiltonian))
            for k in results.keys()
        }
        final_counts = {int_to_bitstring(k, circuit.num_qubits): v for k, v in results.items()}

        depth = circuit.decompose(reps=10).depth()
        if 'cx' in circuit.decompose(reps=10).count_ops():
            cx_count = circuit.decompose(reps=10).count_ops()['cx']
        else:
            cx_count = 0

        return self.construct_result(
            {
                'energy': min(costs),
                'depth': depth,
                'cx_count': cx_count,
                'qpu_time': 0,
                'training_costs': costs,
                'final_sample_energies': final_energies,
                'final_sample_counts': final_counts,
                'optimal_point': opt_params  # needed for educated_guess
            })

    def construct_result(self, result: dict) -> Result:
        counts, energies = result['final_sample_counts'], result['final_sample_energies']
        num_of_samples = sum(counts.values())
        average_energy = statistics.mean(energies.values())
        energy_std = statistics.stdev(energies.values()) if len(energies) > 1 else 0

        best_bs = min(energies, key=energies.get)
        most_common_bs = max(counts, key=counts.get)

        return Result(
            best_bitstring=best_bs,
            best_energy=min(energies.values()),
            most_common_bitstring=most_common_bs,
            most_common_bitstring_energy=energies[most_common_bs],
            distribution={k: v/num_of_samples for k, v in counts.items()},
            energies=energies,
            num_of_samples=num_of_samples,
            average_energy=average_energy,
            energy_std=energy_std,
            result=result
        )


class FALQON(QiskitOptimizationAlgorithm):
    """
    Algorithm class with FALQON.

    Args:
        driver_h (Operator | None): The driver Hamiltonian for the problem.
        delta_t (float): The time step for the evolution operators.
        beta_0 (float): The initial value of beta.
        n (int): The number of iterations to run the algorithm.
        **alg_kwargs: Additional keyword arguments for the base class.

    Attributes:
        driver_h (Operator | None): The driver Hamiltonian for the problem.
        delta_t (float): The time step for the evolution operators.
        beta_0 (float): The initial value of beta.
        n (int): The number of iterations to run the algorithm.
        cost_h (Operator | None): The cost Hamiltonian for the problem.
        n_qubits (int): The number of qubits in the problem.
        parameters (list[str]): The list of algorithm parameters.

    """
    _algorithm_format = 'hamiltonian'

    def __init__(
        self,
        driver_h: SparsePauliOp | None = None,
        delta_t: float = 0.03,
        beta_0: float = 0.0,
        max_reps: int = 20
    ) -> None:
        super().__init__()
        self.driver_h = driver_h
        self.cost_h = None
        self.delta_t = delta_t
        self.beta_0 = beta_0
        self.max_reps = max_reps
        self.n_qubits: int = 0
        self.parameters = ['n', 'delta_t', 'beta_0']

    @property
    def setup(self) -> dict:
        return {
            'driver_h': self.driver_h,
            'delta_t': self.delta_t,
            'beta_0': self.beta_0,
            'n': self.max_reps,
            'cost_h': self.cost_h,
            'n_qubits': self.n_qubits,
            'parameters': self.parameters,
            'arg_kwargs': self.alg_kwargs
        }

    def _get_path(self) -> str:
        return f'{self.name}@{self.max_reps}@{self.delta_t}@{self.beta_0}'

    def run(self, problem: Problem, backend: QiskitBackend, formatter: Callable) -> Result:
        """ Runs the FALQON algorithm """

        if isinstance(backend.sampler, BaseSamplerV1) or isinstance(backend.estimator, BaseEstimatorV1):
            raise ValueError("FALQON works only on V2 samplers and estimators, consider using a different backend.")

        cost_h = formatter(problem)

        if cost_h is None:
            raise ValueError("Formatter returned None")

        self.n_qubits = cost_h.num_qubits

        best_sample, betas, energies, depths, cnot_counts = self._falqon_subroutine(cost_h, backend)

        best_data: BitArray = best_sample[0].data.meas
        counts: dict = best_data.get_counts()
        shots = best_data.num_shots

        result = {'betas': betas,
                  'energies': energies,
                  'depths': depths,
                  'cxs': cnot_counts,
                  'n': self.max_reps,
                  'delta_t': self.delta_t,
                  'beta_0': self.beta_0,
                  'energy': min(energies),
                  }

        return Result(
            best_bitstring=max(counts, key=counts.get),
            most_common_bitstring=max(counts, key=counts.get),
            distribution={k: v/shots for k, v in counts.items()},
            energies=energies,
            energy_std=np.std(energies),
            best_energy=min(energies),
            num_of_samples=shots,
            average_energy=np.mean(energies),
            most_common_bitstring_energy=0,
            result=result
        )

    def _add_ansatz_part(
        self,
        cost_hamiltonian: SparsePauliOp,
        driver_hamiltonian: SparsePauliOp,
        beta: float,
        circuit: QuantumCircuit
    ) -> None:
        """Adds a single FALQON ansatz 'building block' with the specified beta to the circuit"""
        circ_part = QuantumCircuit(circuit.num_qubits)

        circ_part.append(PauliEvolutionGate(cost_hamiltonian, time=self.delta_t), circ_part.qubits)
        circ_part.append(PauliEvolutionGate(driver_hamiltonian, time=self.delta_t * beta), circ_part.qubits)

        circuit.compose(circ_part, circ_part.qubits, inplace=True)

    def _build_ansatz(self, cost_hamiltonian, driver_hamiltonian, betas):
        """Build the FALQON circuit for the given betas"""

        circ = QuantumCircuit(self.n_qubits)
        circ.h(range(self.n_qubits))

        for beta in betas:
            circ.append(PauliEvolutionGate(cost_hamiltonian, time=self.delta_t), circ.qubits)
            circ.append(PauliEvolutionGate(driver_hamiltonian, time=self.delta_t * beta), circ.qubits)
        return circ

    def _falqon_subroutine(
            self,
            cost_hamiltonian: SparsePauliOp,
            backend: QiskitBackend
    ) -> tuple[PrimitiveResult[SamplerPubResult], list[float], list[float], list[int], list[int]]:
        """
        Run the 'meat' of the algorithm.

        Args:
            cost_hamiltonian (SparsePauliOp): Cost hamiltonian from the formatter.
            backend (QiskitBackend): Backend

        Returns:
            tuple[PrimitiveResult[SamplerPubResult], list[float], list[float], list[int], list[int]]:
            Sampler result from best betas, list of betas, list of energies, list of depths, list of cnot counts
        """

        if self.driver_h is None:
            self.driver_h = SparsePauliOp.from_sparse_list([("X", [i], 1) for i in range(self.n_qubits)], num_qubits=self.n_qubits)
            driver_hamiltonian = self.driver_h
        else:
            driver_hamiltonian = self.driver_h

        hamiltonian_commutator = complex(0, 1) * commutator(driver_hamiltonian, cost_hamiltonian)

        betas = [self.beta_0]
        energies = []
        cnot_counts = []
        circuit_depths = []

        circuit = QuantumCircuit(self.n_qubits)
        circuit.h(circuit.qubits)

        self._add_ansatz_part(cost_hamiltonian, driver_hamiltonian, self.beta_0, circuit)

        for i in range(self.max_reps):

            beta = -1 * backend.estimator.run([(circuit, hamiltonian_commutator)]).result()[0].data.evs
            betas.append(beta)

            self._add_ansatz_part(cost_hamiltonian, driver_hamiltonian, beta, circuit)

            energy = backend.estimator.run([(circuit, cost_hamiltonian)]).result()[0].data.evs
            # print(i, energy)
            energies.append(energy)
            circuit_depths.append(circuit.depth())
            cnot_counts.append(circuit.count_ops().get('cx', 0))

        argmin = np.argmin(np.asarray(energies))

        sampling_circuit = self._build_ansatz(cost_hamiltonian, driver_hamiltonian, betas[:argmin])
        sampling_circuit.measure_all()

        best_sample = backend.sampler.run([(sampling_circuit)]).result()

        return best_sample, betas, energies, circuit_depths, cnot_counts


class VQE(QiskitOptimizationAlgorithm):
    """Variational Quantum EigenSolver - qiskit-algorithm implementation wrapper.

    Args:
        optimizer (optimizers.Optimizer | None, optional): Optimizer for VQE. Defaults to None.
        ansatz (QuantumCircuit | None, optional): VQE's ansatz. Defaults to None.
        with_numpy (bool, optional): Ignores ansatz parameter and backend, and changes solver to Numpy based. Defaults to False.
    """
    # pip install git+https://github.com/qiskit-community/qiskit-nature.git
    # pyscf

    def __init__(self, optimizer: optimizers.Optimizer | None = None,
                 ansatz: QuantumCircuit | None = None, with_numpy: bool = False) -> None:
        self.optimizer = optimizers.COBYLA() if optimizer is None else optimizer
        self.ansatz = ansatz
        self.num_qubits: int = 0
        self.with_numpy: bool = with_numpy
        super().__init__()

    @property
    def ansatz(self) -> QuantumCircuit:
        if self._ansatz is None:
            return efficient_su2(self.num_qubits)
        return self._ansatz

    @ansatz.setter
    def ansatz(self, custom_ansatz):
        self._ansatz = custom_ansatz

    def run(self, problem: Problem, backend: Backend, formatter: Callable[..., Any]) -> Result:
        if not isinstance(backend, QiskitBackend):
            raise ValueError('Backend should be QiskitBackend or subclass.')
        if not isinstance(problem, Molecule):
            raise ValueError('The problem for this algorithm should be Molecule problem')
        if not isinstance(problem.operator.num_qubits, int):
            raise ValueError('num_qubits from problem operator is expected to be int')

        estimator = backend.estimator
        self.num_qubits = problem.operator.num_qubits
        if self.with_numpy:
            solver = qiskit_algorithms.NumPyMinimumEigensolver()
        else:
            solver = qiskit_algorithms.VQE(estimator, self.ansatz, self.optimizer)
        vqe_gss = GroundStateEigensolver(problem.mapper, solver)
        vqe_results = vqe_gss.solve(problem.problem)
        return self.construct_result(vqe_results)

    def construct_result(self, result: EigenstateResult) -> Result:
        energy = result.total_energies[0]
        # Not the cleanest way
        return Result(
            '',
            energy,
            '',
            energy,
            {'': 1},
            {'': energy},
            1,
            energy,
            0,
            None
        )



from __future__ import print_function

from bisect import bisect_right

# from pyqubo import Binary
from .Binary import Binary

from .scheduler import JobShopScheduler, KeyList, get_label


def get_jss_bqm(job_dict, max_time, disable_till=None, disable_since=None, disabled_variables=None, lagrange_one_hot=3,
                lagrange_precedence=1, lagrange_share=2):
    if disable_till is None:
        disable_till = {}
    if disable_since is None:
        disable_since = {}
    if disabled_variables is None:
        disabled_variables = []

    scheduler = DWaveScheduler(job_dict, max_time)
    return scheduler.get_bqm(disable_till, disable_since, disabled_variables,
                             lagrange_one_hot,
                             lagrange_precedence,
                             lagrange_share)


class DWaveScheduler(JobShopScheduler):

    def __init__(self, job_dict, max_time=None):
        super().__init__(job_dict, max_time)

    def _add_one_start_constraint(self, lagrange_one_hot=1):
        """self.csp gets the constraint: A task can start once and only once
        """
        for task in self.tasks:
            task_times = {get_label(task, t) for t in range(self.max_time)}
            H_term = 0
            for label in task_times:
                if label in self.absurd_times:
                    continue
                if label not in self.H_vars:
                    var = Binary(label)
                    self.H_vars[label] = var
                else:
                    var = self.H_vars[label]
                H_term += var
            self.H += lagrange_one_hot * ((1 - H_term) ** 2)

    def _add_precedence_constraint(self, lagrange_precedence=1):
        """self.csp gets the constraint: Task must follow a particular order.
         Note: assumes self.tasks are sorted by jobs and then by position
        """
        for current_task, next_task in zip(self.tasks, self.tasks[1:]):
            if current_task.job != next_task.job:
                continue

            # Forming constraints with the relevant times of the next task
            for t in range(self.max_time):
                current_label = get_label(current_task, t)
                if current_label in self.absurd_times:
                    continue

                if current_label not in self.H_vars:
                    var1 = Binary(current_label)
                    self.H_vars[current_label] = var1
                else:
                    var1 = self.H_vars[current_label]

                for tt in range(min(t + current_task.duration, self.max_time)):

                    next_label = get_label(next_task, tt)
                    if next_label in self.absurd_times:
                        continue
                    if next_label not in self.H_vars:
                        var2 = Binary(next_label)
                        self.H_vars[next_label] = var2
                    else:
                        var2 = self.H_vars[next_label]

                    self.H += lagrange_precedence * var1 * var2

    def _add_share_machine_constraint(self, lagrange_share=1):
        """self.csp gets the constraint: At most one task per machine per time unit
        """
        sorted_tasks = sorted(self.tasks, key=lambda x: x.machine)
        # Key wrapper for bisect function
        wrapped_tasks = KeyList(sorted_tasks, lambda x: x.machine)

        head = 0
        while head < len(sorted_tasks):

            # Find tasks that share a machine
            tail = bisect_right(wrapped_tasks, sorted_tasks[head].machine)
            same_machine_tasks = sorted_tasks[head:tail]

            # Update
            head = tail

            # No need to build coupling for a single task
            if len(same_machine_tasks) < 2:
                continue

            # Apply constraint between all tasks for each unit of time
            for task in same_machine_tasks:
                for other_task in same_machine_tasks:
                    if task.job == other_task.job and task.position == other_task.position:
                        continue

                    for t in range(self.max_time):
                        current_label = get_label(task, t)
                        if current_label in self.absurd_times:
                            continue

                        if current_label not in self.H_vars:
                            var1 = Binary(current_label)
                            self.H_vars[current_label] = var1
                        else:
                            var1 = self.H_vars[current_label]

                        for tt in range(t, min(t + task.duration, self.max_time)):
                            this_label = get_label(other_task, tt)
                            if this_label in self.absurd_times:
                                continue
                            if this_label not in self.H_vars:
                                var2 = Binary(this_label)
                                self.H_vars[this_label] = var2
                            else:
                                var2 = self.H_vars[this_label]

                            self.H += lagrange_share * var1 * var2

    def get_bqm(self, disable_till, disable_since, disabled_variables,
                lagrange_one_hot, lagrange_precedence, lagrange_share):
        """Returns a BQM to the Job Shop Scheduling problem.  """

        # Apply constraints to self.csp
        self._remove_absurd_times(
            disable_till, disable_since, disabled_variables)
        self._add_one_start_constraint(lagrange_one_hot)
        self._add_precedence_constraint(lagrange_precedence)
        self._add_share_machine_constraint(lagrange_share)
        # Get BQM
        # bqm = dwaveBinarycsp.stitch(self.csp, **stitch_kwargs)

        # Edit BQM to encourage the shortest schedule
        # Overview of this added penalty:
        # - Want any-optimal-schedule-penalty < any-non-optimal-schedule-penalty
        # - Suppose there are N tasks that need to be scheduled and N > 0
        # - Suppose the the optimal end time for this schedule is t_N
        # - Then the worst optimal schedule would be if ALL the tasks ended at time t_N. (Since
        #   the optimal schedule is only dependent on when the LAST task is run, it is irrelevant
        #   when the first N-1 tasks end.) Note that by "worst" optimal schedule, I am merely
        #   referring to the most heavily penalized optimal schedule.
        #
        # Show math satisfies any-optimal-schedule-penalty < any-non-optimal-schedule-penalty:
        # - Penalty scheme. Each task is given the penalty: base^(task-end-time). The penalty
        #   of the entire schedule is the sum of penalties of these chosen tasks.
        # - Chose the base of my geometric series to be N+1. This simplifies the math and it will
        #   become apparent why it's handy later on.
        #
        # - Comparing the SUM of penalties between any optimal schedule (on left) with that of the
        #   WORST optimal schedule (on right). As shown below, in this penalty scheme, any optimal
        #   schedule penalty <= the worst optimal schedule.
        #     sum_i (N+1)^t_i <= N * (N+1)^t_N, where t_i the time when the task i ends  [eq 1]
        #
        # - Now let's show that all optimal schedule penalties < any non-optimal schedule penalty.
        #   We can prove this by applying eq 1 and simply proving that the worst optimal schedule
        #   penalty (below, on left) is always less than any non-optimal schedule penalty.
        #     N * (N+1)^t_N < (N+1)^(t_N + 1)
        #                               Note: t_N + 1 is the smallest end time for a non-optimal
        #                                     schedule. Hence, if t_N' is the end time of the last
        #                                     task of a non-optimal schedule, t_N + 1 <= t_N'
        #                   <= (N+1)^t_N'
        #                   < sum^(N-1) (N+1)^t_i' + (N+1)^t_N'
        #                   = sum^N (N+1)^t_i'
        #                               Note: sum^N (N+1)^t' is the sum of penalties for a
        #                                     non-optimal schedule
        #
        # - Therefore, with this penalty scheme, all optimal solution penalties < any non-optimal
        #   solution penalties
        base = len(self.last_task_indices) + 1  # Base for exponent
        # Get our pruned (remove_absurd_times) variable list so we don't undo pruning
        # pruned_variables = list(bqm.variables)
        for i in self.last_task_indices:
            task = self.tasks[i]

            for t in range(self.max_time):
                end_time = t + task.duration

                # Check task's end time; do not add in absurd times
                if end_time > self.max_time:
                    continue

                # Add bias to variable
                bias = 2 * base ** (end_time - self.max_time)
                label = get_label(task, t)
                if label in self.absurd_times:
                    continue
                if label not in self.H_vars:
                    var = Binary(label)
                    self.H_vars[label] = var
                else:
                    var = self.H_vars[label]
                self.H += var * bias

        # Get BQM
        self.model = self.H.compile()
        bqm = self.model.to_bqm()
        return bqm



""" Basic problems for Orca """
import numpy as np
from pyqubo import Array
from qlauncher.problems.problem_formulations.jssp.pyqubo_scheduler import get_jss_bqm
import qlauncher.problems.problem_initialization as problem

from qlauncher.base import formatter, adapter


@adapter('qubo', 'qubo_fn')
def qubo_to_fn(qubo):
    # TODO check
    def qubo_fn(bin_vec):
        return np.dot(bin_vec, np.dot(qubo, bin_vec))
    return qubo_fn


@formatter(problem.MaxCut, 'qubo')
def get_maxcut_qubo(problem: problem.MaxCut):
    """ Returns Qubo function """
    n = len(problem.instance)
    Q = np.zeros((n, n))
    for (i, j) in problem.instance.edges:
        Q[i, i] += -1
        Q[j, j] += -1
        Q[i, j] += 1
        Q[j, i] += 1

    return Q, 0


@formatter(problem.EC, 'qubo')
class ECOrca:
    gamma = 1
    delta = 0.05

    def Jrr(self, route1, route2):
        s = len(set(route1).intersection(set(route2)))
        return s / 2

    def hr(self, route1, routes):
        i_sum = 0
        for r in routes:
            i_sum += len(set(r).intersection(set(route1)))
        s = i_sum - len(route1) * 2
        return s / 2

    def calculate_jrr_hr(self, problem: problem.EC):
        Jrr_dict = dict()
        indices = np.triu_indices(len(problem.instance), 1)
        for i1, i2 in zip(indices[0], indices[1]):
            Jrr_dict[(i1, i2)] = self.Jrr(
                problem.instance[i1], problem.instance[i2])

        hr_dict = dict()
        for i in range(len(problem.instance)):
            hr_dict[i] = self.hr(problem.instance[i], problem.instance)

        return Jrr_dict, hr_dict

    def calculate_lengths_tab(self, problem: problem.EC):
        tab = []
        for route in problem.instance:
            tab.append(len(route))
        return tab

    def calculate_num_elements(self, problem: problem.EC):
        d = dict()
        for route in problem.instance:
            for el in route:
                d[el] = 1
        return len(d)

    def calculate_instance_size(self, problem: problem.EC):
        # Calculate instance size for training
        return len(problem.instance)

    def __call__(self, problem: problem.EC):
        self.num_elements = self.calculate_num_elements(problem)
        self.len_routes = self.calculate_lengths_tab(problem)
        Q = np.zeros((len(problem.instance), len(problem.instance)))
        self.Jrr_dict, self.hr_dict = self.calculate_jrr_hr(problem)
        for i in self.Jrr_dict:
            Q[i[0]][i[1]] = self.Jrr_dict[i]
            Q[i[1]][i[0]] = Q[i[0]][i[1]]

        for i in self.hr_dict:
            Q[i][i] = -self.hr_dict[i]

        return Q, 0


@formatter(problem.JSSP, 'qubo')
class JSSPOrca:
    gamma = 1
    lagrange_one_hot = 1
    lagrange_precedence = 2
    lagrange_share = 5

    def _fix_get_jss_bqm(self, instance, max_time, config,
                         lagrange_one_hot=0,
                         lagrange_precedence=0,
                         lagrange_share=0) -> tuple[dict, list, None]:
        pre_result = get_jss_bqm(instance, max_time, config,
                                 lagrange_one_hot=lagrange_one_hot,
                                 lagrange_precedence=lagrange_precedence,
                                 lagrange_share=lagrange_share)
        result = (pre_result.spin.linear, pre_result.spin.quadratic,
                  pre_result.spin.offset)  # I need to change it into dict somehow
        return result, list(result[0].keys()), None

    def calculate_instance_size(self, problem: problem.JSSP):
        # Calculate instance size for training
        _, variables, _ = self._fix_get_jss_bqm(problem.instance, problem.max_time, self.config,
                                                lagrange_one_hot=self.lagrange_one_hot,
                                                lagrange_precedence=self.lagrange_precedence,
                                                lagrange_share=self.lagrange_share)
        return len(variables)

    def get_len_all_jobs(self, problem: problem.JSSP):
        result = 0
        for job in problem.instance.values():
            result += len(job)
        return result

    def one_hot_to_jobs(self, binary_vector, problem: problem.JSSP):
        actually_its_qubo, variables, model = self._fix_get_jss_bqm(problem.instance, problem.max_time, self.config,
                                                                    lagrange_one_hot=self.lagrange_one_hot,
                                                                    lagrange_precedence=self.lagrange_precedence,
                                                                    lagrange_share=self.lagrange_share)
        result = [variables[i]
                  for i in range(len(variables)) if binary_vector[i] == 1]
        return result

    def _set_config(self):
        self.config = {}
        self.config['parameters'] = {}
        self.config['parameters']['job_shop_scheduler'] = {}
        self.config['parameters']['job_shop_scheduler']['problem_version'] = "optimization"

    def __call__(self, problem: problem.JSSP):
        # Define the matrix Q used for QUBO
        self.config = {}
        self.instance_size = self.calculate_instance_size(problem)
        self._set_config()
        actually_its_qubo, variables, model = self._fix_get_jss_bqm(problem.instance, problem.max_time, self.config,
                                                                    lagrange_one_hot=self.lagrange_one_hot,
                                                                    lagrange_precedence=self.lagrange_precedence,
                                                                    lagrange_share=self.lagrange_share)
        reverse_dict_map = {v: i for i, v in enumerate(variables)}

        Q = np.zeros((self.instance_size, self.instance_size))

        for (label_i, label_j), value in actually_its_qubo[1].items():
            i = reverse_dict_map[label_i]
            j = reverse_dict_map[label_j]
            Q[i, j] += value
            Q[j, i] = Q[i, j]

        for label_i, value in actually_its_qubo[0].items():
            i = reverse_dict_map[label_i]
            Q[i, i] += value
        return Q / max(np.max(Q), -np.min(Q)), 0


@formatter(problem.Raw, 'qubo')
def get_raw_qubo(problem: problem.Raw):
    return problem.instance


@formatter(problem.GraphColoring, 'qubo')
def get_graph_coloring_qubo(problem: problem.GraphColoring):
    """ Returns Qubo function """
    num_qubits = problem.instance.number_of_nodes() * problem.num_colors
    x = Array.create('x', shape=(problem.instance.number_of_nodes(), problem.num_colors), vartype='BINARY')
    qubo = 0
    for node in problem.instance.nodes:
        qubo += (1 - sum(x[node, i] for i in range(problem.num_colors))) ** 2
    for n1, n2 in problem.instance.edges:
        for c in range(problem.num_colors):
            qubo += (x[n1, c] * x[n2, c])
    model = qubo.compile()
    qubo_dict, offset = model.to_qubo()
    Q_matrix = np.zeros((num_qubits, num_qubits))
    for i in range(num_qubits):
        for j in range(num_qubits):
            n1, c1 = i//problem.num_colors, i % problem.num_colors
            n2, c2 = j//problem.num_colors, j % problem.num_colors
            key = ("x["+str(n1)+"]["+str(c1)+"]", "x["+str(n2)+"]["+str(c2)+"]")
            if key in qubo_dict:
                Q_matrix[i, j] = qubo_dict[key]
    return Q_matrix, offset



""" Algorithms for qiskit """
from .qiskit_native import QAOA, FALQON
try:
    from .qml import TrainQSVCKernel
except ImportError:
    TrainQSVCKernel = None
try:
    from .educated_guess import EducatedGuess
except ImportError:
    EducatedGuess = None

__all__ = ['QAOA', 'FALQON', 'EducatedGuess', 'TrainQSVCKernel']



""" Hamiltonian formulation of problems """
from itertools import product
import numpy as np

from qiskit import QuantumCircuit
from qiskit.quantum_info import SparsePauliOp
from qiskit_optimization.converters import QuadraticProgramToQubo
from qiskit_optimization.translators import from_ising

from qlauncher.base import formatter
from qlauncher.base.adapter_structure import adapter
import qlauncher.problems.problem_initialization as problems
import qlauncher.hampy as hampy
from qlauncher.hampy import Equation, Variable
from qlauncher.problems.problem_formulations.hamiltonians.tsp import problem_to_hamiltonian as tsp_to_hamiltonian


@adapter('hamiltonian', 'qubo', onehot='quadratic')
def hamiltonian_to_qubo(hamiltonian):
    qp = from_ising(hamiltonian)
    conv = QuadraticProgramToQubo()
    qubo = conv.convert(qp).objective
    return qubo.quadratic.to_array(), 0


@adapter("qubo", "hamiltonian")
def qubo_to_hamiltonian(qubo: np.ndarray) -> SparsePauliOp:
    q_matrix, offset = qubo
    num_vars = q_matrix.shape[0]
    pauli = 0
    for i, col in enumerate(q_matrix):
        for j, entry in enumerate(col):
            if entry == 0:
                continue
            if i == j:
                pauli += SparsePauliOp.from_sparse_list([('I', [0], .5), ('Z', [i], -.5)], num_vars)*entry
            else:
                pauli += SparsePauliOp.from_sparse_list([('I', [0], .25), ('Z', [i], -.25),
                                                        ('Z', [j], -.25), ('ZZ', [i, j], .25)], num_vars)*entry
    pauli += SparsePauliOp.from_sparse_list([('I', [], offset)], num_vars)
    return pauli


def ring_ham(ring: set, n):
    total = None
    ring = list(ring)
    for index in range(len(ring) - 1):
        sparse_list = []
        sparse_list.append((("XX", [ring[index], ring[index + 1]], 1)))
        sparse_list.append((("YY", [ring[index], ring[index + 1]], 1)))
        sp = SparsePauliOp.from_sparse_list(sparse_list, n)
        if total is None:
            total = sp
        else:
            total += sp
    sparse_list = []
    sparse_list.append((("XX", [ring[-1], ring[0]], 1)))
    sparse_list.append((("YY", [ring[-1], ring[0]], 1)))
    sp = SparsePauliOp.from_sparse_list(sparse_list, n)
    total += sp
    return SparsePauliOp(total)


@formatter(problems.EC, 'hamiltonian')
class ECQiskit:
    def __call__(self, problem: problems.EC, onehot='exact') -> SparsePauliOp:
        """ generating hamiltonian"""
        elements = set().union(*problem.instance)
        onehots = []
        for ele in elements:
            ohs = set()
            for i, subset in enumerate(problem.instance):
                if ele in subset:
                    ohs.add(i)
            onehots.append(ohs)
        hamiltonian = None
        for ohs in onehots:
            if onehot == 'exact':
                part = (~hampy.one_in_n(list(ohs), len(problem.instance))).hamiltonian
            elif onehot == 'quadratic':
                part = hampy.one_in_n(list(ohs), len(problem.instance), quadratic=True).hamiltonian

            if hamiltonian is None:
                hamiltonian = part
            else:
                hamiltonian += part
        return hamiltonian.simplify()

    def get_mixer_hamiltonian(self, problem: problems.EC, amount_of_rings=None):
        """ generates mixer hamiltonian """
        def get_main_set():
            main_set = []
            for element_set in problem.instance:
                for elem in element_set:
                    if elem not in main_set:
                        main_set.append(elem)
            return main_set

        def get_constraints():
            constraints, main_set = [], get_main_set()
            for element in main_set:
                element_set = set()
                for index, _ in enumerate(problem.instance):
                    if element in problem.instance[index]:
                        element_set.add(index)
                if len(element_set) > 0 and element_set not in constraints:
                    constraints.append(element_set)

            return constraints

        # creating mixer hamiltonians for all qubits that aren't in rings (in other words applying X gate to them)
        def x_gate_ham(x_gate: list):
            total = None
            for elem in x_gate:
                sparse_list = []
                sparse_list.append((("X", [elem], 1)))
                sp = SparsePauliOp.from_sparse_list(
                    sparse_list, len(problem.instance))
                if total is None:
                    total = sp
                else:
                    total += sp
            return SparsePauliOp(total)

        # looking for all rings in a data and creating a list with them
        ring, x_gate, constraints = [], [], get_constraints()

        ring.append(max(constraints, key=len))

        ring_qubits = set.union(*ring)

        for set_ in constraints:
            if len(set_.intersection(ring_qubits)) == 0:
                ring.append(set_)
                ring_qubits.update(set_)

        if amount_of_rings is not None:
            max_amount_of_rings, user_rings = len(ring), []
            if amount_of_rings > max_amount_of_rings:
                raise ValueError(
                    f"Too many rings. Maximum amount is {max_amount_of_rings}")
            elif amount_of_rings == 0:
                ring_qubits = []
            else:
                current_qubits = ring[0]
                for index in range(amount_of_rings):
                    user_rings.append(ring[index])
                    current_qubits = current_qubits.union(ring[index])
                ring_qubits = current_qubits
        x_gate.extend(id for id, _ in enumerate(
            problem.instance) if id not in ring_qubits)

        # connecting all parts of mixer hamiltonian together
        mix_ham = None
        for set_ in ring:
            if mix_ham is None:
                mix_ham = ring_ham(set_, len(problem.instance))
            else:
                mix_ham += ring_ham(set_, len(problem.instance))

        if mix_ham is None:
            mix_ham = x_gate_ham(x_gate)
        else:
            mix_ham += x_gate_ham(x_gate)

        return mix_ham


@formatter(problems.JSSP, 'hamiltonian')
def get_qiskit_hamiltonian(problem: problems.JSSP) -> SparsePauliOp:
    if problem.optimization_problem:
        return problem.h_o
    else:
        return problem.h_d


@formatter(problems.MaxCut, 'hamiltonian')
def get_qiskit_hamiltonian(problem: problems.MaxCut):
    ham = None
    n = problem.instance.number_of_nodes()
    for edge in problem.instance.edges():
        if ham is None:
            ham = ~hampy.one_in_n(edge, n)
        else:
            ham += ~hampy.one_in_n(edge, n)
    return ham.hamiltonian.simplify()


@formatter(problems.QATM, 'hamiltonian')
class QATMQiskit:
    def __call__(self, problem: problems.QATM, onehot='exact') -> SparsePauliOp:
        cm = problem.instance['cm']
        aircrafts = problem.instance['aircrafts']

        onehot_hamiltonian = None
        for plane, manouvers in aircrafts.groupby(by='aircraft'):
            if onehot == 'exact':
                h = (~hampy.one_in_n(manouvers.index.values.tolist(), len(cm))).hamiltonian
            elif onehot == 'quadratic':
                h = hampy.one_in_n(manouvers.index.values.tolist(), len(cm), quadratic=True).hamiltonian
            elif onehot == 'xor':
                total = None
                eq = Equation(len(cm))
                for part in manouvers.index.values.tolist():
                    if total is None:
                        total = eq[part].to_equation()
                        continue
                    total ^= eq[part]
                h = (~total).hamiltonian
            if onehot_hamiltonian is not None:
                onehot_hamiltonian += h
            else:
                onehot_hamiltonian = h

        triu = np.triu(cm, k=1)
        conflict_hamiltonian = None
        for p1, p2 in zip(*np.where(triu == 1)):
            eq = Equation(len(cm))
            partial_hamiltonian = (eq[int(p1)] & eq[int(p2)]).hamiltonian
            if conflict_hamiltonian is not None:
                conflict_hamiltonian += partial_hamiltonian
            else:
                conflict_hamiltonian = partial_hamiltonian

        hamiltonian = onehot_hamiltonian + conflict_hamiltonian

        if problem.optimization_problem:
            goal_hamiltonian = None
            for i, (maneuver, ac) in problem.instance['aircrafts'].iterrows():
                if maneuver != ac:
                    eq = Equation(len(aircrafts))
                    h = Variable(i, eq).to_equation()
                    if goal_hamiltonian is None:
                        goal_hamiltonian = h
                    else:
                        goal_hamiltonian += h
            goal_hamiltonian /= sum(sum(cm))
            hamiltonian += goal_hamiltonian

        return hamiltonian.simplify()

    def get_mixer_hamiltonian(self, problem: problems.QATM) -> SparsePauliOp:
        cm = problem.instance['cm']
        aircrafts = problem.instance['aircrafts']

        mixer_hamiltonian = None
        for plane, manouvers in aircrafts.groupby(by='aircraft'):
            h = ring_ham(manouvers.index.values.tolist(), len(cm))
            if mixer_hamiltonian is None:
                mixer_hamiltonian = h
            else:
                mixer_hamiltonian += h
        return mixer_hamiltonian

    def get_QAOAAnsatz_initial_state(self, problem: problems.QATM) -> QuantumCircuit:
        aircrafts = problem.instance['aircrafts']
        qc = QuantumCircuit(len(aircrafts))
        for plane, manouvers in aircrafts.groupby(by='aircraft'):
            qc.x(manouvers.index.values.tolist()[0])
        return qc


@formatter(problems.Raw, 'hamiltonian')
def get_qiskit_hamiltonian(problem: problems.Raw) -> SparsePauliOp:
    return problem.instance


@formatter(problems.TSP, 'hamiltonian')
def get_qiskit_hamiltonian(problem: problems.TSP, onehot='exact', constraints_weight=1, costs_weight=1) -> SparsePauliOp:
    return tsp_to_hamiltonian(
        problem,
        onehot=onehot,
        constraints_weight=constraints_weight,
        costs_weight=costs_weight
    )


@formatter(problems.GraphColoring, 'hamiltonian')
def get_qiskit_hamiltonian(problem: problems.GraphColoring, constraints_weight=1, costs_weight=1):
    color_bit_length = int(np.ceil(np.log2(problem.num_colors)))
    num_qubits = problem.instance.number_of_nodes() * color_bit_length
    eq = Equation(num_qubits)
    # Penalty for assigning the same colors to neighboring vertices
    for node1, node2 in problem.instance.edges:
        for ind, comb in enumerate(product(range(2), repeat=color_bit_length)):
            if ind >= problem.num_colors:
                break
            eq2 = None
            for i in range(color_bit_length):
                qubit1 = eq[node1 * color_bit_length + i]
                qubit2 = eq[node2 * color_bit_length + i]
                if comb[i]:
                    exp = qubit1 & qubit2
                else:
                    exp = ~qubit1 & ~qubit2
                if eq2 is None:
                    eq2 = exp
                else:
                    eq2 &= exp
            eq += eq2
    eq *= costs_weight
    # Penalty for using excessive colors
    for node in problem.instance.nodes:
        for ind, comb in enumerate(product(range(2), repeat=color_bit_length)):
            if ind < problem.num_colors:
                continue
            eq2 = None
            for i in range(color_bit_length):
                qubit = eq[node * color_bit_length + i]
                exp = qubit if comb[i] else ~qubit
                if eq2 is None:
                    eq2 = exp
                else:
                    eq2 &= exp
            eq += eq2
    eq *= constraints_weight
    return eq.hamiltonian



from qlauncher.base import Backend
from qlauncher.exceptions import DependencyError

try:
    from ptseries.tbi import create_tbi
except ImportError as e:
    raise DependencyError(e, install_hint='orca', private=True) from e


class OrcaBackend(Backend):
    """ local backend """

    def __init__(self, name: str) -> None:
        super().__init__(name)

    def get_tbi(self):
        """Returns tbi
        """
        return create_tbi()

    def get_args(self):
        return {}



class Binary:
    ''' It's not the same as Binary, it just replicate "spin" calculation "Binary" from pyqubo '''
    def __init__(self, value:str):
        if isinstance(value, str):
            self.data = {value: 0.5}
            self.data['__free__'] = 0.5
        elif isinstance(value, int):
            self.data = {'__free__': value}
        elif isinstance(value, dict):
            self.data = value
        elif isinstance(value, Binary):
            self.data = value.data.copy()
        else:
            raise TypeError(f'Unsupported type: {type(value)}')
        self.variables:list = []
        
    def __str__(self):
        return str(self.data)

    def __add__(self, other):
        data = self.data.copy()
        if isinstance(other, str):
            if other in data:
                data[other] += 1
            else:
                data[other] = 1
        elif isinstance(other, int):
            if '__free__' in data:
                data['__free__'] += other
            else:
                data['__free__'] = other
        elif isinstance(other, float):
            if '__free__' in data:
                data['__free__'] += other
            else:
                data['__free__'] = other
        elif isinstance(other, dict):
            for key, value in other.items():
                if key in data:
                    data[key] += value
                else:
                    data[key] = value
        elif isinstance(other, Binary):
            for key, value in other.data.items():
                if key in data:
                    data[key] += value
                else:
                    data[key] = value
        else:
            raise TypeError(f'Unsupported type: {type(other)}')
        return Binary(data)
    
    def __radd__(self, other):
        data = self.data.copy()
        if isinstance(other, str):
            if other in data:
                data[other] += 1
            else:
                data[other] = 1
        elif isinstance(other, int):
            if '__free__' in data:
                data['__free__'] += other
            else:
                data['__free__'] = other
        elif isinstance(other, float):
            if '__free__' in data:
                data['__free__'] += other
            else:
                data['__free__'] = other
        elif isinstance(other, dict):
            for key, value in other.items():
                if key in data:
                    data[key] += value
                else:
                    data[key] = value
        elif isinstance(other, Binary):
            for key, value in other.data.items():
                if key in data:
                    data[key] += value
                else:
                    data[key] = value
        else:
            raise TypeError(f'Unsupported type: {type(other)}')
        return Binary(data)

    def __sub__(self, other):
        data = self.data.copy()
        if isinstance(other, str):
            if other in data:
                data[other] -= 1
                if data[other] == 0:
                    del data[other]
            else:
                data[other] = -1
        elif isinstance(other, int):
            if '__free__' in data:
                data['__free__'] -= other
                if data['__free__'] == 0:
                    del data['__free__']
            else:
                data['__free__'] = -other
        elif isinstance(other, float):
            if '__free__' in data:
                data['__free__'] -= other
                if data['__free__'] == 0:
                    del data['__free__']
            else:
                data['__free__'] = -other
        elif isinstance(other, dict):
            for key, value in other.items():
                if key in data:
                    data[key] -= value
                    if data[key] == 0:
                        del data[key]
                else:
                    data[key] = -value
        elif isinstance(other, Binary):
            for key, value in other.data.items():
                if key in data:
                    data[key] -= value
                    if data[key] == 0:
                        del data[key]
                else:
                    data[key] = -value
        else:
            raise TypeError(f'Unsupported type: {type(other)}')
        return Binary(data)
    
    def __rsub__(self, other):
        return Binary(other) - self

    def __neg__(self):
        data = {}
        for key, value in self.data.items():
            data[key] = -value
        return Binary(data)

    def __mul__(self, other):
        data = {}
        if isinstance(other, str):
            for key, value in self.data.items():
                if key == other:
                    if '__free__' in data:
                        data['__free__'] += value
                    else:
                        data['__free__'] = value
                else:
                    data[key, other] = value
        elif isinstance(other, int):
            for key, value in self.data.items():
                data[key] = value * other
        elif isinstance(other, float):
            for key, value in self.data.items():
                data[key] = value * other
        elif isinstance(other, dict):
            for key, value in self.data.items():
                for key2, value2 in other.items():
                    data[key, key2] = value * value2
        elif isinstance(other, Binary):
            for key, value in self.data.items():
                for key2, value2 in other.data.items():
                    if key == key2:
                        if '__free__' in data:
                            data['__free__'] += value * value2
                        else:
                            data['__free__'] = value * value2
                    else:
                        data[key, key2] = value * value2
        else:
            raise TypeError(f'Unsupported type: {type(other)}')
        return Binary(data)

    def __rmul__(self, other):
        data = {}
        if isinstance(other, str):
            for key, value in self.data.items():
                if key == other:
                    if '__free__' in data:
                        data['__free__'] += value
                    else:
                        data['__free__'] = value
                else:
                    data[key, other] = value
        elif isinstance(other, int):
            for key, value in self.data.items():
                data[key] = value * other
        elif isinstance(other, float):
            for key, value in self.data.items():
                data[key] = value * other
        elif isinstance(other, dict):
            for key, value in self.data.items():
                for key2, value2 in other.items():
                    data[key, key2] = value * value2
        elif isinstance(other, Binary):
            for key, value in self.data.items():
                for key2, value2 in other.data.items():
                    if key == key2:
                        if '__free__' in data:
                            data['__free__'] += value * value2
                        else:
                            data['__free__'] = value * value2
                    else:
                        data[key, key2] = value * value2
        else:
            raise TypeError(f'Unsupported type: {type(other)}')
        return Binary(data)

    def __pow__(self, other):
        data = self.data.copy()
        new_data = {}
        if isinstance(other, int) and other >= 0:
            if other == 0:
                return Binary({'__free__': 1})
            elif other == 1:
                return Binary(data)
            elif other == 2:
                return Binary(data) * Binary(data)

        else:
            raise TypeError(f'Unsupported type: {type(other)}')
        return Binary(data)

    def __truediv__(self, other):
        return self * (1 / other)

    def calculate(self):

        for key, value in self.data.copy().items():
            if isinstance(key, tuple):
                if key[0] == '__free__':
                    if key[1] in self.data:
                        self.data[key[1]] += value
                    else:
                        self.data[key[1]] = value
                    del self.data[key]
                elif key[1] == '__free__':
                    if key[0] in self.data:
                        self.data[key[0]] += value
                    else:
                        self.data[key[0]] = value
                    del self.data[key]
        energy = 0
        if '__free__' in self.data:
            energy = self.data['__free__']
            del self.data['__free__']

        linear_dict = {}
        quadratic_dict = {}
        for key, value in self.data.items():
            if isinstance(key, str):
                linear_dict[key] = value
            elif isinstance(key, tuple):
                if (key[1], key[0]) in quadratic_dict:
                    quadratic_dict[(key[1], key[0])] += value
                else:
                    quadratic_dict[key] = value
            else:
                raise TypeError(f'Unsupported type: {type(key)}')
        
        # order quadratic
        for key, value in quadratic_dict.copy().items():
            if key[0] > key[1]:
                quadratic_dict[(key[1], key[0])] = value
                del quadratic_dict[key]
        self.variables = list(linear_dict.keys())
        class spin:
            linear = linear_dict
            quadratic = quadratic_dict
            offset = energy
        self.spin = spin
        return linear_dict, quadratic_dict, energy

    def compile(self):
        self.calculate()
        return self

    def to_bqm(self):
        return self



""" Hampy is a small package for creating Hamiltonians from boolean expressions """
from .object import Equation, Variable
from .equations import one_in_n
from .debug import TruthTable
from .utils import shift_affected_qubits

__all__ = ['Equation', 'Variable', 'one_in_n', 'TruthTable', 'shift_affected_qubits']



""" AQT backend class for Qiskit routines """
from typing import Literal
from overrides import override

from qiskit.providers import BackendV1, BackendV2
from qiskit.primitives import Sampler, BaseSamplerV2
from qiskit_ibm_runtime import Options

from qlauncher.routines.qiskit.adapters import SamplerV1ToSamplerV2Adapter
from qlauncher.routines.qiskit import QiskitBackend
from qlauncher.exceptions import DependencyError

try:
    from qiskit_aqt_provider import AQTProvider
    from qiskit_aqt_provider.primitives import AQTSampler, AQTEstimator
except ImportError as e:
    raise DependencyError(e, install_hint='aqt') from e


class AQTBackend(QiskitBackend):
    """
    An extension of QiskitBackend providing support for Alpine Quantum Technologies (AQT) devices.

    Attributes:
        token (str, optional): AQT token, used for authorization when using real device backends.
        dotenv_path (str,optional): (recommended) Path to a .env file containing the AQT token. If dotenv_path is not None, the token will be ignored and the token from the .env file will be used.

    Usage Example
    -------------
    ::

        backend = AQTBackend(token='valid_token', name='device')

    or

    ::

        backend = AQTBackend(dotenv_path='./.env', name='device')

    with a .env file:

    ::

        AQT_TOKEN=valid_token

    """
    sampler: BaseSamplerV2
    estimator: AQTEstimator

    def __init__(
        self,
        name: Literal['local_simulator', 'backendv1v2', 'device'],
        options: Options | None = None,
        backendv1v2: BackendV1 | BackendV2 | None = None,
        auto_transpile_level: Literal[0, 1, 2, 3] | None = None,
        token: str | None = None,
        dotenv_path: str | None = None,
    ) -> None:

        # TODO: This will probably need to be updated to handle custom backend urls, when we get our own computer
        if dotenv_path is None:
            self.provider = AQTProvider(token if token is not None else "DEFAULT_TOKEN", load_dotenv=False)
        else:
            self.provider = AQTProvider(load_dotenv=True, dotenv_path=dotenv_path)
        super().__init__(name, options=options, backendv1v2=backendv1v2, auto_transpile_level=auto_transpile_level)

    @override
    def _set_primitives_on_backend_name(self) -> None:
        if self.name == 'local_simulator':
            self.name = self.provider.backends(backend_type='offline_simulator', name=r".*no_noise")[0].name
        elif self.name == 'backendv1v2':
            if self.backendv1v2 is None:
                raise ValueError("Please indicate a backend when in backendv1v2 mode.")
        elif self.name == 'device':
            available_online_backends = self.provider.backends(backend_type='device')
            if len(available_online_backends) == 0:
                raise ValueError(f"No online backends available for token {self.provider.access_token[:5]}...")
            self.name = available_online_backends[0].name
        else:
            raise ValueError(
                " ".join([
                    f"Unsupported mode for this backend:'{self.name}'."
                    "Please use one of the following: ['local_simulator', 'backendv1v2', 'device']"
                ])
            )

        if self.backendv1v2 is None:
            self.backendv1v2 = self.provider.get_backend(name=self.name)

        self.estimator = AQTEstimator(self.backendv1v2)
        self._samplerV1 = AQTSampler(self.backendv1v2)
        self.sampler = SamplerV1ToSamplerV2Adapter(self._samplerV1)

        self._configure_auto_behavior()



import os
import pickle
import sys

from qlauncher.base.base import Algorithm, Backend, Problem, Result
from qlauncher.launcher.qlauncher import QLauncher
from qlauncher.exceptions import DependencyError
try:
    import dill
    from qcg.pilotjob.api.job import Jobs
    from qcg.pilotjob.api.manager import LocalManager, Manager
except ImportError as e:
    raise DependencyError(e, install_hint='pilotjob') from e


class JobManager:
    def __init__(self, manager: Manager | None = None):
        """
        Job manager is QLauncher's wrapper for process management system, current version works on top of qcg-pilotjob

        Args:
            manager (Manager | None, optional): Manager system to schedule jobs, if set to None, the pilotjob's LocalManager is set.
            Defaults to None.
        """
        self.jobs = {}
        self.code_path = os.path.join(os.path.dirname(__file__), 'pilotjob_task.py')
        if manager is None:
            manager = LocalManager()
        self.manager = manager

    def _count_not_finished(self) -> int:
        return len([job for job in self.jobs.values() if job.get('finished') is False])

    def submit(self, problem: Problem, algorithm: Algorithm, backend: Backend, output_path: str, cores: int | None = None) -> str:
        """
        Submits QLauncher job to the scheduler

        Args:
            problem (Problem): Problem.
            algorithm (Algorithm): Algorithm.
            backend (Backend): Backend.
            output_path (str): Path of output file.
            cores (int | None, optional): Number of cores per task, if None value set to number of free cores (at least 1). Defaults to None.

        Returns:
            str: Job Id.
        """
        if cores is None:
            cores = 1
        job = self._prepare_ql_dill_job(problem=problem, algorithm=algorithm, backend=backend,
                                        output=output_path, cores=cores)
        job_id = self.manager.submit(Jobs().add(**job.get('qcg_args')))[0]
        return job_id

    def submit_many(self, problem: Problem, algorithm: Algorithm, backend: Backend, output_path: str, cores_per_job: int = 1, n_jobs: int | None = None) -> list[str]:
        """
        Submits as many jobs as there are currently available cores.

        Args:
            problem (Problem): Problem.
            algorithm (Algorithm): Algorithm.
            backend (Backend): Backend.
            output_path (str): Path of output file.
            cores_per_job (int, optional): Number of cores per job. Defaults to 1.
            n_jobs: number of jobs to submit. If None, submit as many as possible (free_cores//cores_per_job). Defaults to None.

        Returns:
            list[str]: List with Job Id's.
        """
        free_cores = self.manager.resources()['free_cores']
        if free_cores == 0:
            return []
        qcg_jobs = Jobs()
        for _ in range(n_jobs if n_jobs is not None else free_cores//cores_per_job):
            job = self._prepare_ql_dill_job(problem=problem, algorithm=algorithm, backend=backend, output=output_path, cores=cores_per_job)
            qcg_jobs.add(**job.get('qcg_args'))
        return self.manager.submit(qcg_jobs)

    def wait_for_a_job(self, job_id: str | None = None, timeout: int | float | None = None) -> tuple[str, str]:
        """
        Waits for a job to finish and returns it's id and status.

        Args:
            job_id (str | None, optional): Id of selected job, if None waiting for any job. Defaults to None.
            timeout (int  |  float | None, optional): Timeout in seconds. Defaults to None.

        Raises:
            ValueError: Raises if job_id not found or there are no jobs left.

        Returns:
            tuple[str, str]: job_id, job's status
        """
        if job_id is None:
            if self._count_not_finished() <= 0:
                raise ValueError("There are no jobs left")
            job_id, state = self.manager.wait4_any_job_finish(timeout)
        elif job_id in self.jobs:
            state = self.manager.wait4(job_id, timeout=timeout)[job_id]
        else:
            raise ValueError(f"Job {job_id} not found in {self.__class__.__name__}'s jobs")

        self.jobs[job_id]['finished'] = True
        return job_id, state

    def _prepare_ql_dill_job(self, problem: Problem, algorithm: Algorithm, backend: Backend,
                             output: str, cores: int = 1) -> dict:
        job_uid = f'{len(self.jobs):05d}'
        output_file = os.path.abspath(f'{output}output.{job_uid}')
        launcher = QLauncher(problem, algorithm, backend)
        input_file = os.path.abspath(f'{output}output.{job_uid}')
        with open(input_file, 'wb') as f:
            dill.dump(launcher, f)

        in_args = [self.code_path, input_file, output_file]
        qcg_args = {
            'name': job_uid,
            'exec': sys.executable,
            'args': in_args,
            'model': 'openmpi',
            'stdout': f'{output}stdout.{job_uid}',
            'stderr': f'{output}stderr.{job_uid}',
            'numCores': cores
        }
        job = {'name': job_uid, 'qcg_args': qcg_args, 'output_file': output_file, 'finished': False}
        self.jobs[job_uid] = job
        return job

    def read_results(self, job_id: str) -> Result:
        """
        Reads the result of given job_id.

        Args:
            job_id (str): Job Id.

        Returns:
            Result: Result of selected Job.
        """
        output_path = self.jobs[job_id]['output_file']
        with open(output_path, 'rb') as rt:
            results = pickle.load(rt)
        return results

    def clean_up(self):
        """
        Removes all output files generated in the process and calls self.manager.cleanup().
        """
        for job in self.jobs.values():
            os.remove(job['output_file'])
        if isinstance(self.manager, LocalManager):
            self.manager.cleanup()

    def stop(self):
        """
        Stops the manager process.
        """
        self.manager.cancel(self.jobs)
        if isinstance(self.manager, LocalManager):
            self.manager.finish()

    def __del__(self):
        self.stop()



"""  This module contains the MaxCut class."""
import networkx as nx
import matplotlib.pyplot as plt

from qlauncher.base import Problem


class MaxCut(Problem):
    """
    Class for MaxCut Problem.

    This class represents MaxCut Problem which is a combinatorial optimization problem that involves partitioning the
    vertices of a graph into two sets such that the number of edges between the two sets is maximized. The class contains
    an instance of the problem, so it can be passed into QLauncher.

    Attributes:
        instance (nx.Graph): The graph instance representing the problem.

    """

    def __init__(self, instance: nx.Graph, instance_name='unnamed'):
        super().__init__(instance, instance_name)

    @property
    def setup(self) -> dict:
        return {
            'instance_name': self.instance_name
        }

    @staticmethod
    def from_preset(instance_name: str) -> "MaxCut":
        match instance_name:
            case 'default':
                node_list = list(range(6))
                edge_list = [(0, 1), (0, 2), (0, 5), (1, 3), (1, 4),
                             (2, 4), (2, 5), (3, 4), (3, 5)]
        graph = nx.Graph()
        graph.add_nodes_from(node_list)
        graph.add_edges_from(edge_list)
        return MaxCut(graph, instance_name=instance_name)

    def _get_path(self) -> str:
        return f'{self.name}@{self.instance_name}'

    def visualize(self, bitstring: str | None = None):
        pos = nx.spring_layout(self.instance, seed=42)  # set seed for same node graphs in plt
        plt.figure(figsize=(8, 6))
        if bitstring is None:
            cmap = 'skyblue'
        else:
            cmap = ['crimson' if bit == '1' else 'skyblue' for bit in bitstring]
        nx.draw(self.instance, pos, with_labels=True, node_color=cmap,
                node_size=500, edge_color='gray', font_size=10, font_weight='bold')
        plt.title("Max-Cut Problem Instance Visualization")
        plt.show()

    @staticmethod
    def generate_maxcut_instance(num_vertices: int, edge_probability: float) -> "MaxCut":
        graph = nx.gnp_random_graph(num_vertices, edge_probability)
        return MaxCut(graph, instance_name='Generated')



"""DWave algorithms"""

from collections.abc import Callable

from qlauncher.base import Algorithm, Problem, Backend, Result
from qlauncher.exceptions import DependencyError
from qlauncher.routines.dwave.backends import BQMBackend
try:
    from dimod.binary.binary_quadratic_model import BinaryQuadraticModel
    from dimod import SampleSet
except ImportError as e:
    raise DependencyError(e, install_hint='dwave') from e


class DwaveSolver(Algorithm):
    _algorithm_format = 'bqm'

    def __init__(self, chain_strength=1, num_reads=1000, **alg_kwargs) -> None:
        self.chain_strength = chain_strength
        self.num_reads = num_reads
        self.label: str = 'TBD_TBD'
        super().__init__(**alg_kwargs)

    def run(self, problem: Problem, backend: Backend, formatter: Callable) -> Result:
        if not isinstance(backend, BQMBackend):
            raise ValueError(f'{backend.__class__} is not supported by DwaveSolver algorithm, use BQMBackend instead')
        self.label = f'{problem.name}_{problem.instance_name}'

        bqm: BinaryQuadraticModel = formatter(problem)

        res = self._solve_bqm(bqm, backend.sampler, **self.alg_kwargs)
        return self._construct_result(res)

    def _solve_bqm(self, bqm, sampler, **kwargs):
        res = sampler.sample(
            bqm, num_reads=self.num_reads, label=self.label, chain_strength=self.chain_strength, **kwargs)
        return res

    def _construct_result(self, result: SampleSet) -> Result:
        distribution = {}
        energies = {}
        for (value, energy, occ) in zip(result.record.sample, result.record.energy, result.record.num_occurrences, strict=True):
            bitstring = ''.join(map(str, value))
            if bitstring in distribution:
                distribution[bitstring] += occ
                continue
            distribution[bitstring] = occ
            energies[bitstring] = energy

        return Result.from_distributions(distribution, energies, result)



from __future__ import print_function

from bisect import bisect_right

import qlauncher.hampy as hampy

from .scheduler import JobShopScheduler, KeyList, get_label


def get_jss_hamiltonian(job_dict, max_time, onehot):
    scheduler = QiskitScheduler(job_dict, max_time, onehot)
    return scheduler.get_hamiltonian()


class QiskitScheduler(JobShopScheduler):

    def __init__(self, job_dict, max_time=None, onehot='exact'):
        super().__init__(job_dict, max_time)
        self.H_pos_by_label = dict()
        self.H_label_by_pos = dict()
        self.onehot = onehot

    def _add_one_start_constraint(self, lagrange_one_hot=1):
        for task in self.tasks:
            task_times = {get_label(task, t) for t in range(self.max_time)}
            onehot_tasks = set()
            for label in task_times:
                if label in self.absurd_times:
                    continue
                else:
                    onehot_tasks.add(self.H_pos_by_label[label])
            if self.onehot == 'exact':
                self.H += (~hampy.one_in_n(onehot_tasks, self.n)).hamiltonian
            elif self.onehot == 'quadratic':
                self.H += hampy.one_in_n(onehot_tasks, self.n, quadratic=True).hamiltonian

    def _add_precedence_constraint(self, lagrange_precedence=1):
        for current_task, next_task in zip(self.tasks, self.tasks[1:]):
            if current_task.job != next_task.job:
                continue
            for t in range(self.max_time):
                current_label = get_label(current_task, t)
                if current_label in self.absurd_times:
                    continue
                var1 = self.H_pos_by_label[current_label]
                for tt in range(min(t + current_task.duration, self.max_time)):
                    next_label = get_label(next_task, tt)
                    if next_label in self.absurd_times:
                        continue
                    var2 = self.H_pos_by_label[next_label]
                    equation = hampy.Equation(self.n)
                    self.H += (equation[var1] & equation[var2]).hamiltonian

    def _add_share_machine_constraint(self, lagrange_share=1):
        sorted_tasks = sorted(self.tasks, key=lambda x: x.machine)
        wrapped_tasks = KeyList(sorted_tasks, lambda x: x.machine)

        head = 0
        while head < len(sorted_tasks):

            tail = bisect_right(wrapped_tasks, sorted_tasks[head].machine)
            same_machine_tasks = sorted_tasks[head:tail]

            head = tail

            if len(same_machine_tasks) < 2:
                continue

            for task in same_machine_tasks:
                for other_task in same_machine_tasks:
                    if task.job == other_task.job and task.position == other_task.position:
                        continue

                    for t in range(self.max_time):
                        current_label = get_label(task, t)
                        if current_label in self.absurd_times:
                            continue

                        var1 = self.H_pos_by_label[current_label]

                        for tt in range(t, min(t + task.duration, self.max_time)):
                            this_label = get_label(other_task, tt)
                            if this_label in self.absurd_times:
                                continue
                            var2 = self.H_pos_by_label[this_label]
                            equation = hampy.Equation(self.n)
                            self.H += (equation[var1] & equation[var2]).hamiltonian

    def _build_variable_dict(self):
        for task in self.tasks:
            task_times = {get_label(task, t) for t in range(self.max_time)}
            for label in task_times:
                if label in self.absurd_times:
                    continue
                else:
                    self.H_pos_by_label[label] = len(self.H_pos_by_label)
                    self.H_label_by_pos[len(self.H_label_by_pos)] = label
        self.n = len(self.H_pos_by_label)

    def get_hamiltonian(self):
        self._remove_absurd_times({}, {}, [])
        self._build_variable_dict()
        self._add_one_start_constraint()
        self._add_precedence_constraint()
        self._add_share_machine_constraint()
        # Get BQM
        # bqm = dwavebinarycsp.stitch(self.csp, **stitch_kwargs)

        # Edit BQM to encourage the shortest schedule
        # Overview of this added penalty:
        # - Want any-optimal-schedule-penalty < any-non-optimal-schedule-penalty
        # - Suppose there are N tasks that need to be scheduled and N > 0
        # - Suppose the the optimal end time for this schedule is t_N
        # - Then the worst optimal schedule would be if ALL the tasks ended at time t_N. (Since
        #   the optimal schedule is only dependent on when the LAST task is run, it is irrelevant
        #   when the first N-1 tasks end.) Note that by "worst" optimal schedule, I am merely
        #   referring to the most heavily penalized optimal schedule.
        #
        # Show math satisfies any-optimal-schedule-penalty < any-non-optimal-schedule-penalty:
        # - Penalty scheme. Each task is given the penalty: base^(task-end-time). The penalty
        #   of the entire schedule is the sum of penalties of these chosen tasks.
        # - Chose the base of my geometric series to be N+1. This simplifies the math and it will
        #   become apparent why it's handy later on.
        #
        # - Comparing the SUM of penalties between any optimal schedule (on left) with that of the
        #   WORST optimal schedule (on right). As shown below, in this penalty scheme, any optimal
        #   schedule penalty <= the worst optimal schedule.
        #     sum_i (N+1)^t_i <= N * (N+1)^t_N, where t_i the time when the task i ends  [eq 1]
        #
        # - Now let's show that all optimal schedule penalties < any non-optimal schedule penalty.
        #   We can prove this by applying eq 1 and simply proving that the worst optimal schedule
        #   penalty (below, on left) is always less than any non-optimal schedule penalty.
        #     N * (N+1)^t_N < (N+1)^(t_N + 1)
        #                               Note: t_N + 1 is the smallest end time for a non-optimal
        #                                     schedule. Hence, if t_N' is the end time of the last
        #                                     task of a non-optimal schedule, t_N + 1 <= t_N'
        #                   <= (N+1)^t_N'
        #                   < sum^(N-1) (N+1)^t_i' + (N+1)^t_N'
        #                   = sum^N (N+1)^t_i'
        #                               Note: sum^N (N+1)^t' is the sum of penalties for a
        #                                     non-optimal schedule
        #
        # - Therefore, with this penalty scheme, all optimal solution penalties < any non-optimal
        #   solution penalties

        # Get BQM
        decision_hamiltonian = self.H.simplify().copy()

        base = len(self.last_task_indices) + 1  # Base for exponent
        # Get our pruned (remove_absurd_times) variable list so we don't undo pruning
        # pruned_variables = list(bqm.variables)

        h = 0

        for i in self.last_task_indices:
            task = self.tasks[i]

            for t in range(self.max_time):
                end_time = t + task.duration

                # Check task's end time; do not add in absurd times
                if end_time > self.max_time:
                    continue

                # Add bias to variable
                bias = 2 * base ** (end_time - self.max_time)
                # bias = base**(end_time - 5)
                # bias = base ** (end_time)
                label = get_label(task, t)
                if label in self.absurd_times:
                    continue

                var = self.H_pos_by_label[label]
                self.H += hampy.Variable(var, hampy.Equation(self.n)).to_equation().hamiltonian
        # self.H += h / (base * len(self.last_task_indices))

        # Get BQM
        optimization_hamiltonian = self.H.simplify().copy()
        return decision_hamiltonian, optimization_hamiltonian, self.H_pos_by_label.copy(), self.H_label_by_pos.copy()



# TODO: reimplement
# from qiskit_optimization.converters import QuadraticProgramToQubo
# from qiskit_optimization.translators import from_ising
# from qiskit.quantum_info import SparsePauliOp
# import numpy as np

# from qlauncher.exceptions import qubo_to_hamiltonian, _qubo_dict_into_hamiltonian, _qubo_matrix_into_hamiltonian


# def _hamiltonian_to_qubo(hamiltonian: SparsePauliOp) -> tuple[np.ndarray, float]:
#     qp = from_ising(hamiltonian)
#     conv = QuadraticProgramToQubo()
#     qubo = conv.convert(qp).objective
#     return qubo.quadratic.to_array(), qubo.constant


# def test_qubo_dict_to_hamiltonian():
#     qubo, offset = {('a', 'a'): 1, ('a', 'b'): 2, ('b', 'b'): 3}, 2
#     hamiltonian = _qubo_dict_into_hamiltonian(qubo, offset=offset)
#     assert hamiltonian.num_qubits == 2
#     new_qubo, new_offset = _hamiltonian_to_qubo(hamiltonian)
#     assert new_offset == offset
#     qubo_matrix = [[1, 2], [0, 3]]
#     assert (new_qubo == qubo_matrix).all()



from typing import TypeVar
from qiskit import QuantumCircuit, transpile
from qiskit.providers.backend import BackendV1, BackendV2
from qiskit.primitives import BackendSamplerV2, BackendEstimatorV2
from qiskit_ibm_runtime import EstimatorV2, SamplerV2


AUTO_TRANSPILE_SAMPLER_TYPE = TypeVar('AUTO_TRANSPILE_SAMPLER_TYPE', BackendSamplerV2, SamplerV2)
AUTO_TRANSPILE_ESTIMATOR_TYPE = TypeVar('AUTO_TRANSPILE_ESTIMATOR_TYPE', BackendEstimatorV2, EstimatorV2)


def _get_transpiled_sampler_pubs(pubs: list[tuple | QuantumCircuit], backend: BackendV1 | BackendV2, optimization_level: int = 2) -> list[tuple | QuantumCircuit]:
    new_pubs = []
    for pub in pubs:
        if isinstance(pub, QuantumCircuit):
            pub = transpile(pub, backend, optimization_level=optimization_level)
        elif isinstance(pub, tuple):
            pub = (transpile(pub[0], backend, optimization_level=optimization_level), *pub[1:])
        new_pubs.append(pub)
    return new_pubs


def _get_transpiled_estimator_pubs(pubs: list[tuple], backend: BackendV1 | BackendV2, optimization_level: int = 2) -> list[tuple | QuantumCircuit]:
    new_pubs = []
    for pub in pubs:
        circuit, operator, *args = pub
        transp_circ: QuantumCircuit = transpile(circuit, backend, optimization_level=optimization_level)
        transp_op = operator.apply_layout(transp_circ.layout, num_qubits=transp_circ.num_qubits)
        pub = (transp_circ, transp_op, *args)
        new_pubs.append(pub)
    return new_pubs


def _assign_sampler_pubs(pubs: list[tuple | QuantumCircuit]) -> list[QuantumCircuit]:
    new_pubs = []
    for pub in pubs:
        if isinstance(pub, tuple):
            qc, *params = pub
            pub = qc.assign_parameters(*params)
        new_pubs.append(pub)
    return new_pubs


def _assign_estimator_pubs(pubs: list[tuple]) -> list[tuple]:
    new_pubs = []
    for pub in pubs:
        if len(pub) > 2:
            qc, observable, *params = pub
            pub = (qc.assign_parameters(*params), observable)
        new_pubs.append(pub)
    return new_pubs


def set_sampler_auto_run_behavior(sampler: AUTO_TRANSPILE_SAMPLER_TYPE, auto_transpile: bool = False, auto_transpile_level: int = 2, auto_assign: bool = False) -> AUTO_TRANSPILE_SAMPLER_TYPE:
    """
    Set chosen automatic behavior on a sampler instance.

    Args:
        sampler (AUTO_TRANSPILE_SAMPLER_TYPE): Compatible sampler instance
        auto_transpile (bool, optional): Whether to automatically transpile to the samplers backend. Defaults to False.
        auto_transpile_level (int, optional): What level of optimization to set. Defaults to 2.
        auto_assign (bool, optional): Whether to automatically assign parameters in case of parameterized circuits. Defaults to False.

    Returns:
        AUTO_TRANSPILE_SAMPLER_TYPE: Same instance with modified run() method.
    """
    func = sampler.run

    def run_wrapper(pubs, *args, shots: int | None = None):
        if auto_transpile:
            pubs = _get_transpiled_sampler_pubs(pubs, sampler._backend, optimization_level=auto_transpile_level)
        if auto_assign:
            pubs = _assign_sampler_pubs(pubs)

        return func(pubs, *args, shots=shots)

    sampler.run = run_wrapper
    return sampler


def set_estimator_auto_run_behavior(estimator: AUTO_TRANSPILE_ESTIMATOR_TYPE, auto_transpile: bool = False, auto_transpile_level: int = 2, auto_assign: bool = False) -> AUTO_TRANSPILE_ESTIMATOR_TYPE:
    """
    Set chosen automatic behavior on a estimator instance.

    Args:
        estimator (AUTO_TRANSPILE_ESTIMATOR_TYPE): Compatible estimator instance
        auto_transpile (bool, optional): Whether to automatically transpile to the estimators backend. Defaults to False.
        auto_transpile_level (int, optional): What level of optimization to set. Defaults to 2.
        auto_assign (bool, optional): Whether to automatically assign parameters in case of parameterized circuits. Defaults to False.

    Returns:
        AUTO_TRANSPILE_ESTIMATOR_TYPE: Same instance with modified run() method.
    """
    func = estimator.run

    def run_wrapper(pubs, *args, precision: float | None = None):
        if auto_transpile:
            pubs = _get_transpiled_estimator_pubs(pubs, estimator._backend, optimization_level=auto_transpile_level)
        if auto_assign:
            pubs = _assign_estimator_pubs(pubs)
        return func(pubs, *args, precision=precision)
    estimator.run = run_wrapper
    return estimator



"""
Routine file for Cirq library
"""
from typing import Any, Iterable, Literal
from collections.abc import Sequence
import math

import numpy as np
import qiskit

from qiskit.result import QuasiDistribution
from qiskit.primitives import Sampler, BaseSamplerV2, BitArray, DataBin
from qiskit.primitives.primitive_job import PrimitiveJob
from qiskit.primitives.base.base_primitive_job import BasePrimitiveJob
from qiskit.primitives.base.sampler_result import SamplerResult
from qiskit.primitives.containers.primitive_result import PrimitiveResult
from qiskit.primitives.containers.sampler_pub_result import SamplerPubResult
from qiskit.primitives.containers.sampler_pub import SamplerPubLike

from qlauncher.base import Backend
from qlauncher.base.translator import Translation
from qlauncher.exceptions import DependencyError
try:
    import cirq
    from cirq.sim.sparse_simulator import Simulator
except ImportError as e:
    raise DependencyError(e, install_hint='cirq') from e


def extract_bitstrings_from_result(
    result: cirq.Result
) -> list[str]:
    measurements = result.measurements

    sorted_keys = list(measurements.keys())

    bitstrings = []
    num_shots = len(measurements[sorted_keys[0]])

    for shot_index in range(num_shots):
        bits = []
        for key in sorted_keys:
            bits.extend(str(b) for b in measurements[key][shot_index])

        bitstring = "".join(bits)
        bitstrings.append(bitstring)
    return bitstrings


def cirq_result_to_counts(
    result: cirq.Result
) -> dict:
    bitstrings = extract_bitstrings_from_result(result)

    counts = {}
    for bs in bitstrings:
        counts[bs] = counts.get(bs, 0) + 1
    return counts


def cirq_result_to_probabilities(
    result: cirq.Result,
    integer_keys: bool = False
) -> dict:
    counts = cirq_result_to_counts(result)

    total_shots = sum(counts.values())
    if integer_keys:
        prob_dict = {int(k, 2): v / total_shots for k, v in counts.items()}
    else:
        prob_dict = {k: v / total_shots for k, v in counts.items()}

    return prob_dict


class _CirqRunner:
    simulator = Simulator()
    repetitions = 1024

    @classmethod
    def calculate_circuit(
            cls,
            circuit: qiskit.QuantumCircuit,
            return_type: Literal['counts', 'dist', 'raw'] = 'counts',
            shots: int | None = None
    ) -> dict | list[str]:

        if circuit.num_clbits == 0:
            circuit = circuit.measure_all(inplace=False)

        cirq_circ = Translation.get_translation(circuit, 'cirq')

        result = cls.simulator.run(cirq_circ, repetitions=cls.repetitions if shots is None else shots)

        if return_type == 'raw':
            return extract_bitstrings_from_result(result)

        return cirq_result_to_counts(result) if return_type == 'counts' else cirq_result_to_probabilities(result)


class CirqSampler(Sampler):
    """ Sampler adapter for Cirq """

    def _call(self, circuits: Sequence[int], parameter_values: Sequence[Sequence[float]], **run_options) -> SamplerResult:
        bound_circuits = []
        for i, value in zip(circuits, parameter_values):
            bound_circuits.append(
                self._circuits[i]
                if len(value) == 0
                else self._circuits[i].assign_parameters(dict(zip(self._parameters[i], value)))
            )
        distributions = [_CirqRunner.calculate_circuit(circuit, 'dist') for circuit in bound_circuits]
        quasi_dists = list(map(QuasiDistribution, distributions))
        return SamplerResult(quasi_dists, [{} for _ in range(len(parameter_values))])


class CirqSamplerV2(BaseSamplerV2):
    def __init__(self) -> None:
        super().__init__()
        self.cirq_sampler_v1 = CirqSampler()

    def _run(self, bound_circuits, shots):
        bitstring_collections = [_CirqRunner.calculate_circuit(circuit, 'raw', shots) for circuit in bound_circuits]

        results = []
        for bsc in bitstring_collections:
            arr = np.array(
                [
                    np.frombuffer(int(bs, 2).to_bytes(math.ceil(len(bs)/8)), dtype=np.uint8)
                    for bs in bsc
                ])
            bit_array = BitArray(
                arr,
                num_bits=len(bsc[0])
            )
            results.append(SamplerPubResult(data=DataBin(meas=bit_array), metadata={'shots': len(bitstring_collections[0])}))
        return PrimitiveResult(results, metadata={'version': 2})

    def run(
        self,
        pubs: Iterable[SamplerPubLike],
        *,
        shots: int | None = None
    ) -> BasePrimitiveJob[PrimitiveResult[SamplerPubResult], Any]:
        bound_circuits = []
        for pub in pubs:
            if isinstance(pub, qiskit.QuantumCircuit):
                bound_circuits.append(pub)
            elif len(pub) == 1 and isinstance(pub[0], qiskit.QuantumCircuit):
                bound_circuits.append(pub[0])
            elif len(pub) == 2:
                bound_circuits.append(pub[0].assign_parameters(pub[1]))
            else:
                raise ValueError(
                    f"Incorrect pub, expected QuantumCircuit, tuple[QuantumCircuit] or tuple[QuantumCircuit, Iterable[float]], got {type(pub)}")

        job = PrimitiveJob(self._run, bound_circuits, shots)
        job._submit()
        return job


class CirqBackend(Backend):
    """

    Args:
        Backend (_type_): _description_
    """

    def __init__(self, name: Literal['local_simulator'] = 'local_simulator'):
        self.sampler = CirqSamplerV2()
        self.samplerV1 = CirqSampler()
        super().__init__(name)



"""
``qiskit``
================

The QLauncher version for Qiskit-based architecture.
"""
from .algorithms import QAOA, EducatedGuess, FALQON, TrainQSVCKernel
from qlauncher.routines.qiskit.backends.qiskit_backend import QiskitBackend
from qlauncher.routines.qiskit.backends.ibm_backend import IBMBackend
from qlauncher.routines.qiskit.backends.aqt_backend import AQTBackend
from qlauncher.routines.qiskit.backends.aer_backend import AerBackend
from qlauncher.problems.problem_formulations.hamiltonian import *



import numpy as np
from qlauncher.base import Problem


class TabularML(Problem):
    def __init__(self, X: np.ndarray, y: np.ndarray | None = None, instance_name: str = 'unnamed') -> None:
        super().__init__((X, y), instance_name)

    @property
    def setup(self) -> dict:
        X, y = self.instance
        return {
            'X_shape': X.shape,
            'y_shape': y.shape if y is not None else None,
            'instance_name': self.instance_name
        }

    @staticmethod
    def from_preset(instance_name: str, **kwargs) -> 'TabularML':
        match instance_name:
            case 'default':
                return TabularML(
                    np.array([
                        [5.2,  3.1],
                        [11.3,  2.2],
                        [9.8, 10.5],
                        [2.1,  1.9],
                        [12.0, 15.2],
                        [6.4,  9.9],
                        [0.0, 11.1],
                        [10.1, 10.2],
                        [8.8,  7.7],
                        [13.3,  0.4]
                    ]),
                    np.array([
                        0,
                        1,
                        1,
                        0,
                        1,
                        0,
                        1,
                        1,
                        0,
                        1
                    ]),
                    instance_name=instance_name
                )
        raise ValueError()

    def visualize(self):
        X, y = self.instance
        print(f"Tabular ML problem '{self.name}'")
        print(X[:10])
        print(y[:10] if y is not None else 'No target variable')



import json
import os
import sys
from qlauncher import QLauncher
from qlauncher.routines.qiskit import QAOA, QiskitBackend
from qlauncher.problems import MaxCut, EC, JSSP, QATM, Problem
import dill

PROBLEM_DICT: dict[str, type[Problem]] = {
    'MaxCut': MaxCut,
    'EC': EC,
    'JSSP': JSSP,
    'QATM': QATM
}

ALGORITHM_DICT = {
    'QAOA': QAOA,
}

BACKEND_DICT = {
    'QiskitBackend': QiskitBackend
}


def parse_arguments() -> tuple[QLauncher, str]:
    """ Returns QLauncher object and output file path """
    if len(sys.argv) == 3:
        input_file_path = sys.argv[1]
        with open(input_file_path, 'rb') as f:
            launcher = dill.load(f)
        os.remove(input_file_path)
        output_path = sys.argv[2]
    elif len(sys.argv) == 6:
        problem = PROBLEM_DICT[sys.argv[1]]
        algorithm = ALGORITHM_DICT[sys.argv[2]]
        backend = BACKEND_DICT[sys.argv[3]]
        kwargs = json.loads(sys.argv[4])
        launcher = QLauncher(problem(**kwargs.get('problem', dict())),
                             algorithm(**kwargs.get('algorithm', dict())),
                             backend(**kwargs.get('backend', dict())))
        output_path = sys.argv[5]
    else:
        raise ValueError(f'Wrong number of arguments, expected 3 or 6 got {len(sys.argv)} instead')

    return launcher, output_path


def main():
    launcher, output_path = parse_arguments()

    launcher.run()
    launcher.save(path=output_path, save_format='pickle')


if __name__ == '__main__':
    main()



from typing import overload
from qiskit.quantum_info import SparsePauliOp


class Equation:
    @overload
    def __init__(self, size: int): ...
    @overload
    def __init__(self, hamiltonian: SparsePauliOp): ...
    @overload
    def __init__(self, sparse_list: list[tuple], size: int): ...

    def __init__(self, argument, *args):
        if isinstance(argument, int):
            self.size = argument
            self._hamiltonian = SparsePauliOp.from_sparse_list([('I', [], 0)], argument)
        elif isinstance(argument, SparsePauliOp):
            self.size = argument.num_qubits
            self._hamiltonian = argument
        elif isinstance(argument, list) and len(args) > 0 and isinstance(args[0], int):
            self.size = args[0]
            self._hamiltonian = SparsePauliOp.from_sparse_list(argument, args[0])
        else:
            raise TypeError('Wrong arguments!')

    def get_variable(self, index: int) -> "Variable":
        assert isinstance(index, int), "Index needs to be an integer"
        obj = Variable(index, self)
        return obj

    @property
    def hamiltonian(self) -> SparsePauliOp:
        return self._hamiltonian.simplify()

    @hamiltonian.setter
    def hamiltonian(self, new_hamiltonian: SparsePauliOp):
        self._hamiltonian = new_hamiltonian

    def to_sparse_pauli_op(self) -> SparsePauliOp:
        return self.hamiltonian

    def get_order(self) -> int:
        equation_order = 0
        for Z_term in self.hamiltonian.paulis:
            equation_order = max(equation_order, str(Z_term).count('Z'))
        return equation_order

    def is_quadratic(self) -> bool:
        return all(term.z.sum() <= 2 for term in self.hamiltonian.paulis)

    def __or__(self, other: "Variable | Equation", /) -> "Equation":
        if isinstance(other, Variable):
            other = other.to_equation()

        return Equation(self.hamiltonian + other.hamiltonian - self.hamiltonian.compose(other.hamiltonian))

    def __and__(self, other: "Variable | Equation", /) -> "Equation":
        if isinstance(other, Variable):
            other = other.to_equation()

        return Equation(self.hamiltonian.compose(other.hamiltonian))

    def __xor__(self, other: "Variable | Equation", /) -> "Equation":
        if isinstance(other, Variable):
            other = other.to_equation()

        return Equation(self.hamiltonian + other.hamiltonian - (2 * self.hamiltonian.compose(other.hamiltonian)))

    def __invert__(self) -> "Equation":
        I = ('I', [], 1)
        identity = SparsePauliOp.from_sparse_list([I], self.size)
        return Equation(identity - self.hamiltonian)

    def __getitem__(self, variable_number: int):
        return self.get_variable(variable_number)

    def __eq__(self, other: "Equation") -> bool:
        if isinstance(other, Variable):
            other = other.to_equation()

        return self.hamiltonian == other.hamiltonian

    def __add__(self, other: "Variable | Equation") -> "Equation":
        if isinstance(other, Variable):
            other = other.to_equation()

        return Equation(self.hamiltonian + other.hamiltonian)

    def __radd__(self, other: "Equation") -> "Equation":
        if isinstance(other, Variable):
            other = other.to_equation()

        return Equation(self.hamiltonian + other.hamiltonian)

    def __mul__(self, other: "Equation | float") -> "Equation":
        if isinstance(other, Variable):
            other = other.to_equation()
        if isinstance(other, float) or isinstance(other, int):
            return Equation(float(other) * self.hamiltonian)
        return Equation(self.hamiltonian.compose(other.hamiltonian))

    def __rmul__(self, other: "Equation | float") -> "Equation":
        if isinstance(other, Variable):
            other = other.to_equation()
        if isinstance(other, float) or isinstance(other, int):
            return Equation(float(other) * self.hamiltonian)
        return Equation(self.hamiltonian.compose(other.hamiltonian))


class Variable:
    def __init__(self, index: int, parent: Equation):
        self.index = index
        self.size = parent.size

    def __xor__(self, other: "Equation | float", /) -> Equation:
        if isinstance(other, Equation):
            return self.to_equation() ^ other

        I = ('I', [], 0.5)
        Z_term = ('ZZ', [self.index, other.index], -0.5)
        eq = Equation(SparsePauliOp.from_sparse_list([I, Z_term], self.size))
        return eq

    def __or__(self, other: "Variable | Equation", /) -> Equation:
        if isinstance(other, Equation):
            return self.to_equation() | other

        I_term = ('I', [], 0.75)
        Z1_term = ('Z', [self.index], -0.25)
        Z2_term = ('Z', [other.index], -0.25)
        ZZ_term = ('ZZ', [self.index, other.index], -0.25)
        eq = Equation([I_term, Z1_term, Z2_term, ZZ_term], self.size)
        return eq

    def __and__(self, other: "Variable | Equation", /) -> Equation:

        if isinstance(other, Equation):
            return self.to_equation() & other

        I_term = ('I', [], 0.25)
        Z1_term = ('Z', [self.index], -0.25)
        Z2_term = ('Z', [other.index], -0.25)
        ZZ_term = ('ZZ', [self.index, other.index], 0.25)
        eq = Equation([I_term, Z1_term, Z2_term, ZZ_term], self.size)
        return eq

    def __invert__(self) -> Equation:
        I_term = ('I', [], 0.5)
        Z_term = ('Z', [self.index], 0.5)
        return Equation([I_term, Z_term], self.size)

    def to_equation(self) -> Equation:
        I_term = ('I', [], 0.5)
        Z_term = ('Z', [self.index], -0.5)
        return Equation([I_term, Z_term], self.size)



"""
QLauncher
================

Program made for easier usage of Quantum computer for testing purposes.
Provided with pre-made problems and algorithms.
"""
from qlauncher.base import Result
from qlauncher.launcher import QLauncher



from collections.abc import Sequence
from typing import Any, Iterable

import math

import numpy as np

from qiskit import transpile, QuantumCircuit
from qiskit.primitives.containers.primitive_result import PrimitiveResult
from qiskit.primitives.containers.sampler_pub_result import SamplerPubResult
from qiskit.result import QuasiDistribution
from qiskit.primitives import SamplerResult, BasePrimitiveJob, BitArray, DataBin
from qiskit.primitives.base import BaseSamplerV1, BaseSamplerV2
from qiskit.primitives.containers.sampler_pub import SamplerPubLike
from qiskit.primitives.primitive_job import PrimitiveJob


class RuntimeJobV2Adapter(BasePrimitiveJob):
    def __init__(self, job, **kwargs):
        super().__init__(job.job_id(), **kwargs)
        self.job = job

    def result(self):
        raise NotImplementedError()

    def cancel(self):
        return self.job.cancel()

    def status(self):
        return self.job.status()

    def done(self):
        return self.job.done()

    def cancelled(self):
        return self.job.cancelled()

    def running(self):
        return self.job.running()

    def in_final_state(self):
        return self.job.in_final_state()


class SamplerV2JobAdapter(RuntimeJobV2Adapter):
    """
    Dummy data holder, returns a v1 SamplerResult from v2 sampler job.
    """

    def __init__(self, job, **kwargs):
        super().__init__(job, **kwargs)

    def _get_quasi_meta(self, res):
        data = BitArray.concatenate_bits(list(res.data.values()))
        counts = data.get_int_counts()
        probs = {k: v/data.num_shots for k, v in counts.items()}
        quasi_dists = QuasiDistribution(probs, shots=data.num_shots)

        metadata = res.metadata
        metadata["sampler_version"] = 2  # might be useful for debugging

        return quasi_dists, metadata

    def result(self):
        res = self.job.result()
        qd, metas = [], []
        for r in res:
            quasi_dist, metadata = self._get_quasi_meta(r)
            qd.append(quasi_dist)
            metas.append(metadata)

        return SamplerResult(quasi_dists=qd, metadata=metas)


def _transpile_circuits(circuits, backend):
    # Transpile qaoa circuit to backend instruction set, if backend is provided
    # ? I pass a backend into SamplerV2 as *mode* but here sampler_v2.mode returns None, why?
    if not backend is None:
        if isinstance(circuits, Sequence):
            circuits = [transpile(circuit) for circuit in circuits]
        else:
            circuits = transpile(circuits)

    return circuits


class SamplerV2ToSamplerV1Adapter(BaseSamplerV1):
    """
    Adapts a v2 sampler to a v1 interface.
    """

    def __init__(self, sampler_v2: BaseSamplerV2, backend=None):
        """
        Args:
            sampler_v2 (BaseSamplerV2): V2 sampler to be adapted.
            backend (Backend | None): Backend to transpile circuits to.
        """
        self.sampler_v2 = sampler_v2
        self.backend = backend
        super().__init__()

    def _run(self, circuits, parameter_values=None, **run_options) -> SamplerV2JobAdapter:
        circuits = _transpile_circuits(circuits, self.backend)
        v2_list = list(zip(circuits, parameter_values))
        job = self.sampler_v2.run(pubs=v2_list, **run_options)

        return SamplerV2JobAdapter(job)


class SamplerV1ToSamplerV2Adapter(BaseSamplerV2):
    """
    Adapts a v1 sampler to a v2 interface.

    Args:
        BaseSamplerV2 (_type_): _description_
    """

    def __init__(self, sampler_v1: BaseSamplerV1) -> None:
        super().__init__()
        self.samplerv1 = sampler_v1

    def _run(self, pubs: Iterable[SamplerPubLike], shots: int = 1024):
        circuits, params = [], []
        for pub in pubs:
            if isinstance(pub, QuantumCircuit):
                circuits.append(pub)
                params.append([])
            elif isinstance(pub, tuple):
                circuits.append(pub[0])
                params.append(pub[1] if len(pub) == 2 else [])

        out = self.samplerv1.run(circuits, params, shots=shots).result()
        results = []
        for dist in out.quasi_dists:
            vals: list[int] = []
            max_val = 0
            for k, v in dist.items():
                vals += [k] * int(round(v*shots, 0))
                max_val = max(max_val, k)

            required_bits = math.ceil(math.log2(max_val))
            required_bytes = math.ceil(required_bits/8)

            arr = np.array(
                [
                    np.frombuffer(v.to_bytes(required_bytes), dtype=np.uint8)
                    for v in vals
                ])
            bit_array = BitArray(
                arr,
                num_bits=required_bits
            )
            results.append(SamplerPubResult(data=DataBin(meas=bit_array), metadata={'shots': shots}))

        return PrimitiveResult(results, metadata={'version': 2})

    def run(
        self,
        pubs: Iterable[SamplerPubLike],
        *,
        shots: int | None = None
    ) -> BasePrimitiveJob[PrimitiveResult[SamplerPubResult], Any]:
        job = PrimitiveJob(self._run, pubs, shots if shots is not None else 1024)
        job._submit()
        return job



from typing import Literal
from collections.abc import Callable
import numpy as np

from qlauncher.base import Problem, Algorithm, Backend, Result
from qlauncher.exceptions import DependencyError
from qlauncher.routines.orca.backends import OrcaBackend

try:
    from ptseries.algorithms.binary_solvers import BinaryBosonicSolver
except ImportError as e:
    raise DependencyError(e, install_hint='orca', private=True) from e


class BBS(Algorithm):
    """
    Binary Bosonic Solver algorithm class.

    This class represents the Binary Bosonic Solver (BBS) algorithm. BBS is a quantum-inspired algorithm that
    solves optimization problems by mapping them onto a binary bosonic system. It uses a training process
    to find the optimal solution.

    ### Attributes:

    - algorithm_format ('qubo', 'fn'), optional): If the algorithm input is a function or a qubo matrix. Defaults to 'qubo'.
    - input_state (list[int] | None, optional): Photonic circuit input state provided to the ORCA computer. If None defaults to [1,0,1,0,1...]. Defaults to None.
    - n_samples (int, optional): Number of samples. Defaults to 100.
    - gradient_mode (str, optional): Gradient mode. Defaults to "parameter-shift".
    - gradient_delta (float, optional): Gradient Delta. Defaults to np.pi/6.
    - sampling_factor (int, optional): Number of times quantum samples are passed through the classical flipping layer. Defaults to 1.
    - learning_rate (float, optional): Learning rate of the algorithm. Defaults to 5e-2.
    - learning_rate_flip (float, optional): Bit flip learning rate. Defaults to 1e-1.
    - updates (int, optional): Number of epochs. Defaults to 100.

    """
    _algorithm_format = 'qubo'

    def __init__(self, algorithm_format: Literal['qubo', 'fn'] = 'qubo',
                 input_state: list[int] | None = None,
                 n_samples: int = 100,
                 gradient_mode: str = "parameter-shift",
                 gradient_delta: float = np.pi / 6,
                 sampling_factor: int = 1,
                 learning_rate: float = 5e-2,
                 learning_rate_flip: float = 1e-1,
                 updates: int = 100):
        super().__init__()
        self._algorithm_format = algorithm_format
        self.bbs_params = {
            'n_samples': n_samples,
            'gradient_mode': gradient_mode,
            'gradient_delta': gradient_delta,
            'sampling_factor': sampling_factor,
        }
        self.training_params = {
            'learning_rate': learning_rate,
            'learning_rate_flip': learning_rate_flip,
            'updates': updates,
        }
        self.input_state = input_state

    def run(self, problem: Problem, backend: Backend, formatter: Callable[[Problem], np.ndarray]) -> Result:

        if not isinstance(backend, OrcaBackend):
            raise ValueError(f'{backend.__class__} is not supported by BBS algorithm, use OrcaBackend instead')
        objective = formatter(problem)

        # TODO: use offset somehow
        if not callable(objective):
            objective, offset = objective

        if self.input_state is None:
            if not callable(objective):
                self.input_state = [(i + 1) % 2 for i in range(len(objective))]
            else:
                raise ValueError('input_state needs to be provided if objective is a function (callable)')

        tbi = backend.get_tbi()
        bbs = BinaryBosonicSolver(pb_dim=len(self.input_state),
                                  objective=objective,
                                  input_state=self.input_state,
                                  tbi=tbi,
                                  **self.bbs_params)

        bbs.solve(**self.training_params)

        return self.construct_results(bbs)

    def get_bitstring(self, result: list[float]) -> str:
        return ''.join(map(str, map(int, result)))

    def construct_results(self, solver: BinaryBosonicSolver) -> Result:
        # TODO: add support for distribution (probably with different logger)
        best_bitstring = ''.join(
            map(str, map(int, solver.config_min_encountered)))
        best_energy = solver.E_min_encountered
        most_common_bitstring = None
        most_common_bitstring_energy = None
        distribution = None
        energy = None
        num_of_samples = solver.n_samples
        average_energy = None
        energy_std = None
        #! Todo: instead of None attach relevant info from 'results'
        # results fail to pickle correctly btw
        return Result(best_bitstring, best_energy, most_common_bitstring,
                      most_common_bitstring_energy, distribution, energy,
                      num_of_samples, average_energy, energy_std, None)



""" Algorithms for Qiskit routines """
import math
import os
from collections.abc import Callable

import weakref

import numpy as np
import scipy

from qlauncher.base import Problem, Algorithm, Result
from qlauncher.routines.qiskit.backends.qiskit_backend import QiskitBackend
from qlauncher.workflow.pilotjob_scheduler import JobManager
from qlauncher.routines.qiskit.algorithms.qiskit_native import QAOA


class EducatedGuess(Algorithm):
    _algorithm_format = 'hamiltonian'

    def __init__(self, starting_p: int = 3, max_p: int = 8, max_job_batch_size: int | None = None, verbose: bool = False):
        """
        Algorithm utilizing all available cores to run multiple QAOA's in parallel to find optimal parameters.

        Args:
            starting_p (int, optional): Initial value of QAOA's p parameter. Defaults to 3.
            max_p (int, optional): Maximum value for QAOA's p parameter. Defaults to 8.
            max_job_batch_size: Maximum number of jobs to run for a given p value. If None, run as many as possible. Defaults to None.
            verbose (bool, optional): Verbose. Defaults to False.
        """
        self.output_initial = 'initial/'
        self.output_interpolated = 'interpolated/'
        self.output = 'output/'
        self.p_init = starting_p
        self.p_max = max_p
        self.verbose = verbose
        self.failed_jobs = 0
        self.min_energy = math.inf
        self.manager = JobManager()
        self.best_job_id = ''
        self.max_jobs = max_job_batch_size

        weakref.finalize(self, self.manager.stop)  # Kill the running jobs in case of a crash or otherwise

    def run(self, problem: Problem, backend: QiskitBackend, formatter: Callable) -> Result:
        self.manager.submit_many(problem, QAOA(p=self.p_init), backend, output_path=self.output_initial, n_jobs=self.max_jobs)
        print(f'{len(self.manager.jobs)} jobs submitted to qcg')

        found_optimal_params = False

        while not found_optimal_params:
            jobid, state = self.manager.wait_for_a_job()

            if state != 'SUCCEED':
                self.failed_jobs += 1
                continue
            has_potential, energy = self._process_job(jobid, self.p_init, self.min_energy, compare_factor=0.99)
            if has_potential:
                found_optimal_params = self._search_for_job_with_optimal_params(jobid, energy, problem, backend)

            self.manager.submit_many(problem, QAOA(p=self.p_init), backend, output_path=self.output_initial, n_jobs=self.max_jobs)

        result = self.manager.read_results(self.best_job_id)
        self.manager.stop()
        return result

    def _search_for_job_with_optimal_params(self, previous_job_id, previous_energy, problem, backend) -> bool:
        new_job_id = None
        for p in range(self.p_init + 1, self.p_max + 1):
            previous_job_results = self.manager.read_results(previous_job_id).result
            initial_point = self._interpolate_f(list(previous_job_results['result']['optimal_point']), p-1)

            new_job_id = self.manager.submit(problem, QAOA(p=p, initial_point=initial_point),
                                             backend, output_path=self.output_interpolated)
            _, state = self.manager.wait_for_a_job(new_job_id)
            if state != 'SUCCEED':
                self.failed_jobs += 1
                return False
            has_potential, new_energy = self._process_job(new_job_id, p, previous_energy)
            if has_potential:
                previous_energy = new_energy
                previous_job_id = new_job_id
            else:
                return False
        self.best_job_id = new_job_id if new_job_id is not None else previous_job_id
        return True

    def _process_job(self, jobid: str, p: int, energy_to_compare: float, compare_factor: float = 1.0) -> tuple[
            float, bool]:
        result = self.manager.read_results(jobid).result
        optimal_point = result['optimal_point']
        has_potential = False
        linear = self._check_linearity(optimal_point, p)
        energy = result['energy']
        if self.verbose:
            print(f'job {jobid}, p={p}, energy: {energy}')

        if p == self.p_init and energy < energy_to_compare:
            print(f'new min energy: {energy}')
            self.min_energy = energy
        if linear and energy * compare_factor < energy_to_compare:
            has_potential = True
        return has_potential, energy

    def _create_directories_if_not_existing(self):
        if not os.path.exists(self.output_initial):
            os.makedirs(self.output_initial)
        if not os.path.exists(self.output_interpolated):
            os.makedirs(self.output_interpolated)
        if not os.path.exists(self.output):
            os.makedirs(self.output)

    def _interp(self, params: np.ndarray) -> np.ndarray:
        arr1 = np.append([0], params)
        arr2 = np.append(params, [0])
        weights = np.arange(len(arr1)) / len(params)
        res = arr1 * weights + arr2 * weights[::-1]
        return res

    def _interpolate_f(self, params: np.ndarray, p: int) -> np.ndarray:
        betas = params[:p]
        gammas = params[p:]
        new_betas = self._interp(betas)
        new_gammas = self._interp(gammas)
        return np.hstack([new_betas, new_gammas])

    def _check_linearity(self, optimal_params: np.ndarray, p: int) -> bool:
        linear = False
        correlations = (scipy.stats.pearsonr(np.arange(1, p + 1), optimal_params[:p])[0],
                        scipy.stats.pearsonr(np.arange(1, p + 1), optimal_params[p:])[0])

        if abs(correlations[0]) > 0.85 and abs(correlations[1]) > 0.85:
            linear = True
        return linear



"""qiskit_aer implementation of QiskitBackend"""
from typing import Literal

from qiskit.providers import BackendV1, BackendV2
from qiskit.primitives import BackendSamplerV2, BackendEstimatorV2
from qiskit_ibm_runtime import Options
from qiskit_aer import AerSimulator
from qiskit_aer.noise import NoiseModel

from qlauncher.routines.qiskit.backends.qiskit_backend import QiskitBackend


class AerBackend(QiskitBackend):
    """
    QiskitBackend utilizing the qiskit_aer library. Runs local simulations only, utilizing CUDA capable gpus if available.
    """

    def __init__(
        self,
        name: Literal['local_simulator', 'backendv1v2'],
        options: Options | None = None,
        backendv1v2: BackendV1 | BackendV2 | None = None,
        auto_transpile_level: Literal[0, 1, 2, 3] | None = None,
        simulation_method: str = 'automatic',
        simulation_device: Literal['CPU', 'GPU'] = 'CPU',
    ) -> None:
        self.method = simulation_method
        self.device = simulation_device
        super().__init__(name, options, backendv1v2, auto_transpile_level)

    def _set_primitives_on_backend_name(self):
        if self.name == 'local_simulator':
            self.simulator = AerSimulator(method=self.method, device=self.device)
        elif self.name == 'backendv1v2':
            if self.backendv1v2 is None:
                raise AttributeError(
                    'Please indicate a backend when in backendv1v2 mode.')
            noise_model = NoiseModel.from_backend(self.backendv1v2)
            self.simulator = AerSimulator(method=self.method, device=self.device, noise_model=noise_model)
        else:
            raise ValueError(
                f"Unsupported mode for this backend:'{self.name}'. Please use one of the following: ['local_simulator', 'backendv1v2']")

        self.sampler = BackendSamplerV2(backend=self.simulator)
        self.estimator = BackendEstimatorV2(backend=self.simulator)

        self._configure_auto_behavior()

    def set_options(self, **fields):
        """Set additional options for the instance AerSimulator"""
        self.simulator.set_options(**fields)



""" Module providing Translation between different universal quantum computers """
from typing import Any
from abc import ABC, abstractmethod

import qiskit
from qiskit import qasm2
from qiskit.compiler import transpile
from qiskit.transpiler.passes import RemoveBarriers

from qlauncher.exceptions import DependencyError
try:
    import cirq
    from cirq.contrib.qasm_import.qasm import circuit_from_qasm
except ImportError as e:
    raise DependencyError(e, 'cirq') from e


class Translation(ABC):
    """ Translation layer for circuits written in different languages """
    basis_gates: list[str] = []
    language: str = 'None'
    circuit_object: type = type
    language_name_mapping: dict[str, "Translation"] = {}
    circuit_class_mapping: dict[type, "Translation"] = {}

    def __init_subclass__(cls):
        Translation.language_name_mapping[cls.language] = cls()
        Translation.circuit_class_mapping[cls.circuit_object] = cls()

    @abstractmethod
    def to_qasm(self, circuit: Any) -> str:
        """ Translation from given circuit into qasm (as a string) """

    @abstractmethod
    def from_qasm(self, qasm: str) -> Any:
        """ Translation given in qasm (as a string) circuit into language specific object """

    @staticmethod
    def get_translation(circuit: Any, language: str) -> Any:
        """ Transpiles circuit into given languages basis_gates, translates it to qasm, and from qasm into desired languages object. """
        circuit_qasm_translator = Translation.circuit_class_mapping[circuit.__class__]
        qasm_circuit_translator = Translation.language_name_mapping[language]
        if isinstance(circuit, qiskit.QuantumCircuit):
            transpiled_circuit = transpile_circuit(circuit, qasm_circuit_translator.basis_gates)
        else:
            transpiled_circuit = circuit  # Transpilation is usually not needed
        qasm = circuit_qasm_translator.to_qasm(transpiled_circuit)
        return qasm_circuit_translator.from_qasm(qasm)


class CirqTranslation(Translation):
    """ Translation class for Cirq """
    basis_gates = ['x', 'y', 'z', 'cx', 'h', 'rx', 'ry', 'rz']
    language: str = 'cirq'
    circuit_object = cirq.Circuit

    def to_qasm(self, circuit: cirq.Circuit) -> str:
        return circuit.to_qasm()

    def from_qasm(self, qasm: str) -> cirq.Circuit:
        return circuit_from_qasm(qasm)


class QiskitTranslation(Translation):
    """ Translation class for Qiskit """
    basis_gates = ['x', 'y', 'z', 'cx', 'h', 'rx', 'ry', 'rz', 'u']
    language: str = 'qiskit'
    circuit_object = qiskit.QuantumCircuit

    def to_qasm(self, circuit: qiskit.QuantumCircuit) -> str:
        return qasm2.dumps(circuit)

    def from_qasm(self, qasm: str) -> qiskit.QuantumCircuit:
        return qiskit.QuantumCircuit.from_qasm_str(qasm)


def transpile_circuit(qc: qiskit.QuantumCircuit, basis_gates: list[str]) -> qiskit.QuantumCircuit:
    """Makes circuit compatible with cirq

    Args:
        qc (qiskit.QuantumCircuit): circuit

    Returns:
        qiskit.QuantumCircuit: transpiled circuit
    """
    qc_transpiled = transpile(
        qc,
        basis_gates=basis_gates,
        optimization_level=3
    )

    return RemoveBarriers()(qc_transpiled)



from qlauncher.problems.problem_initialization import TabularML
from qlauncher.base import formatter


@formatter(TabularML, 'tabular_ml')
def tabular_formatter(problem: TabularML):
    return problem.instance



"""  This module contains the EC class."""
import ast
from collections import defaultdict

import networkx as nx
import matplotlib.pyplot as plt

from qlauncher.base import Problem


class EC(Problem):
    """
    Class for exact cover problem.

    The exact cover problem is a combinatorial optimization problem that involves finding a subset of a given set
    of elements such that the subset covers all elements and the number of elements in the subset is minimized.
    The class contains an instance of the problem, so it can be passed into QLauncher.

    Attributes:
        onehot (str): The one-hot encoding used for the problem.
        instance (any): The instance of the problem.
        instance_name (str | None): The name of the instance.
        instance_path (str): The path to the instance file.

    """

    def __init__(self, instance: list[set[int]] = None, instance_name: str = 'unnamed') -> None:
        super().__init__(instance=instance, instance_name=instance_name)

    @property
    def setup(self) -> dict:
        return {
            'instance_name': self.instance_name
        }

    def _get_path(self) -> str:
        return f'{self.name}@{self.instance_name}'

    @staticmethod
    def from_preset(instance_name: str, **kwargs) -> "EC":
        match instance_name:
            case 'micro':
                instance = [{1, 2},
                            {1}]
            case 'default':
                instance = [{1, 4, 7},
                            {1, 4},
                            {4, 5, 7},
                            {3, 5, 6},
                            {2, 3, 6, 7},
                            {2, 7}]
        return EC(instance=instance, instance_name=instance_name, **kwargs)

    @classmethod
    def from_file(cls, path: str, **kwargs) -> "EC":
        with open(path, 'r', encoding='utf-8') as file:
            read_file = file.read()
        instance = ast.literal_eval(read_file)
        return EC(instance, **kwargs)

    def read_instance(self, instance_path: str):
        with open(instance_path, 'r', encoding='utf-8') as file:
            read_file = file.read()
        self.instance = ast.literal_eval(read_file)

    def visualize(self, marked: str | None = None):
        G = nx.Graph()
        size = len(self.instance)
        ec = list(map(lambda x: set(map(str, x)), self.instance))
        names = set.union(*ec)
        for i in range(len(ec)):
            G.add_node(i)
        for i in sorted(names):
            G.add_node(i)
        covered = defaultdict(int)
        for node, edges in enumerate(ec):
            for goal_node in edges:
                G.add_edge(node, goal_node)

        edge_colors = []
        for goal_node in G.edges():
            node, str_node = goal_node
            if marked is None:
                edge_colors.append('black')
                continue
            if marked[node] == '1':
                edge_colors.append('red')
                covered[str_node] += 1
            else:
                edge_colors.append('gray')
        color_map = []
        for node in G:
            if isinstance(node, int):
                color_map.append('lightblue')
            elif covered[node] == 0:
                color_map.append('yellow')
            elif covered[node] == 1:
                color_map.append('lightgreen')
            else:
                color_map.append('red')
        pos = nx.bipartite_layout(G, nodes=range(size))
        nx.draw(G, pos, node_color=color_map, with_labels=True,
                edge_color=edge_colors)
        plt.title('Exact Cover Problem Visualization')
        plt.show()

    @staticmethod
    def generate_ec_instance(n: int, m: int, p: float = 0.5, **kwargs) -> "EC":
        graph = nx.bipartite.random_graph(n, m, p)
        right_nodes = [n for n, d in graph.nodes(data=True) if d["bipartite"] == 0]
        instance = [set() for _ in right_nodes]
        for left, right in graph.edges():
            instance[left].add(right)
        return EC(instance=instance, **kwargs)



""" Module with functionalities for debugging Hamiltonians and checking their boolean properties """
from itertools import product
import numpy as np
import matplotlib.pyplot as plt
from qiskit.quantum_info import SparsePauliOp

from qlauncher.hampy.object import Equation


class TruthTable:
    def __init__(self, equation: Equation | SparsePauliOp, return_int: bool = True):
        if isinstance(equation, SparsePauliOp):
            hamiltonian = equation
            self.size = hamiltonian.num_qubits
        if isinstance(equation, Equation):
            hamiltonian = equation.hamiltonian
            self.size = equation.size
        self.return_int = return_int
        self.truth_table = self._ham_to_truth(hamiltonian)
        self.lowest_value = min(self.truth_table.values())

    def count(self, value: float) -> int:
        return list(self.truth_table.values()).count(value)

    def get_solutions(self, value: float) -> list[str]:
        return list(filter(lambda x: self.truth_table[x] == value, self.truth_table.keys()))

    def count_min_value_solutions(self) -> int:
        return self.count(self.lowest_value)

    def get_min_value_solutions(self) -> list[str]:
        return self.get_solutions(self.lowest_value)

    def check_if_binary(self) -> bool:
        return all((value == 0 or value == 1) for value in self.truth_table.values())

    def plot_distribution(self) -> None:
        values = list(self.truth_table.values())
        counts, bins = np.histogram(values, max(values)+1)
        plt.stairs(counts, bins)
        plt.show()

    def _ham_to_truth(self, hamiltonian: SparsePauliOp):
        return {
            ''.join(reversed(bitstring)): value for bitstring, value in
            zip(
                product(('0', '1'), repeat=self.size),
                map(lambda x: int(x.real), hamiltonian.to_matrix().diagonal())
                if self.return_int else
                hamiltonian.to_matrix().diagonal()
            )
        }

    def __getitem__(self, index: str | int):
        if isinstance(index, int):
            index = bin(index)[2:].zfill(self.size)
        return self.truth_table[index]



"""Wrapper for QLauncher that enables the user to launch tasks asynchronously (futures + multiprocessing)"""
from typing import Literal
from collections.abc import Callable
import time

from qlauncher.problems import Raw
from qlauncher.base.base import Backend, Algorithm, Problem, Result
from qlauncher.launcher.qlauncher import QLauncher
from qlauncher.launcher.aql.aql_task import AQLTask, get_timeout


class AQL:
    """
    Launches QLauncher task asynchronously.

    Attributes:
        tasks (list[AQLTask]): list of submitted tasks.
        mode (Literal[&#39;default&#39;, &#39;optimize_session&#39;]): Execution mode.

    Usage Example
    -------------
    ::

        aql = AQL(mode='optimize_session')

        t_real = aql.add_task(
            (
                GraphColoring.from_preset('small'), 
                QAOA(),
                AQTBackend(token='valid_token', name='device')
            ),
            constraints_weight=5,
            costs_weight=1
            )

        aql.add_task(
            (
                GraphColoring.from_preset('small'), 
                QAOA(),
                QiskitBackend('local_simulator')
            ),
            dependencies=[t_real],
            constraints_weight=5,
            costs_weight=1
        )

        aql.start()
        result_real, result_sim = aql.results(timeout=15)


    """

    def __init__(
        self,
        mode: Literal['default', 'optimize_session'] = 'default'
    ) -> None:
        """
        Args:
            mode (Literal[&#39;default&#39;, &#39;optimize_session&#39;], optional): 
                    Task execution mode. 
                    If 'optimize_session' all tasks running on a real quantum device get split into separate generation and run subtasks,
                    then the quantum tasks are ran in one shorter block. 
                    Defaults to 'default'.
        """

        self.tasks: list[AQLTask] = []
        self.mode: Literal['default', 'optimize_session'] = mode

        self._classical_tasks: list[AQLTask] = []
        self._quantum_tasks: list[AQLTask] = []

    def running_task_count(self) -> int:
        """
        Returns:
            int: Amount of tasks that are currently executing.
        """
        return sum(1 for t in self.tasks if t.running())

    def cancel_running_tasks(self):
        """Cancel all running tasks assigned to this AQL instance."""
        for t in self._classical_tasks + self._quantum_tasks:
            t.cancel()

    def results(self, timeout: float | int | None = None, cancel_tasks_on_timeout: bool = True) -> list[Result | None]:
        """
        Get a list of results from tasks.
        Results are ordered in the same way the tasks were added.
        Blocks the thread until all tasks are finished.

        Args:
            timeout (float | int | None, optional): 
                    The maximum amount to wait for execution to finish. 
                    If None, wait forever. If not None and time runs out, raises TimeoutError. 
                    Defaults to None.
            cancel_tasks_on_timeout (bool): Whether to cancel all other tasks when one task raises a TimeoutError.

        Returns:
            list[Result | None]: Task results.
        """
        try:
            start = time.time()
            return [t.result(timeout=get_timeout(timeout, start)) if not t.cancelled() else None for t in self.tasks]
        except TimeoutError as e:
            if cancel_tasks_on_timeout:
                self.cancel_running_tasks()
            raise e
        except Exception as e:
            self.cancel_running_tasks()
            raise e

    def add_task(
        self,
        launcher: QLauncher | tuple[Problem, Algorithm, Backend],
        dependencies: list[AQLTask] | None = None,
        callbacks: list[Callable] | None = None,
        **kwargs
    ) -> AQLTask:
        """
        Add a QLauncher task to the execution queue.

        Args:
            launcher (QLauncher | tuple[Problem, Algorithm, Backend]): Launcher instance that will be run.
            dependencies (list[AQLTask] | None, optional): Tasks that must finish first before this task. Defaults to None.
            callbacks (list[Callable] | None, optional): 
                    Functions to run when the task finishes. 
                    The task will be passed to the function as the only parameter. 
                    Defaults to None.

        Returns:
            AQLTask: Pointer to the submitted task.
        """
        if isinstance(launcher, tuple):
            launcher = QLauncher(*launcher)

        dependencies_list = dependencies if dependencies is not None else []

        if self.mode != 'optimize_session' or not launcher.backend.is_device:
            task = AQLTask(
                lambda: launcher.run(**kwargs),
                dependencies=dependencies,
                callbacks=(callbacks if callbacks is not None else [])
            )
            self.tasks.append(task)
            self._classical_tasks.append(task)
            return task

        def gen_task():
            launcher.formatter.set_run_params(kwargs)
            return launcher.formatter(launcher.problem)

        # Split real device task into generation and actual run on a QC
        t_gen = AQLTask(
            gen_task,
            dependencies=[dep for dep in dependencies_list if not dep in self._quantum_tasks]
        )

        def quantum_task(formatted, *_):
            ql = QLauncher(
                Raw(formatted, launcher.problem.instance_name),
                launcher.algorithm,
                launcher.backend
            )
            return ql.run()

        t_quant = AQLTask(
            quantum_task,
            dependencies=[t_gen] + [dep for dep in dependencies_list if dep in self._quantum_tasks],
            callbacks=(callbacks if callbacks is not None else []),
            pipe_dependencies=True  # Receive output from formatter
        )

        self._classical_tasks.append(t_gen)
        self._quantum_tasks.append(t_quant)
        self.tasks.append(t_quant)

        return t_quant

    def start(self):
        """Start tasks execution."""
        for t in self._classical_tasks+self._quantum_tasks:
            if t.cancelled() or t.done() or t.running():
                raise ValueError("Cannot start again, some tasks were already ran or are currently running.")

        self._run_async()

    def _run_async(self):
        quantum_dependencies = set()
        dependency_queue = self._quantum_tasks.copy()
        while dependency_queue:
            t = dependency_queue.pop(0)
            quantum_dependencies |= set(t.dependencies)
            dependency_queue += t.dependencies

        quantum_dependencies = quantum_dependencies.difference(self._quantum_tasks)
        # The gateway tasks will ensure execution order of
        # (all classical tasks that quantum tasks depend on) - (all quantum tasks) - (rest of the classical tasks)
        gateway_task_classical = AQLTask(
            lambda: True,
            dependencies=list(quantum_dependencies)
        )

        for qt in self._quantum_tasks:
            qt.dependencies.append(gateway_task_classical)

        gateway_task_quantum = AQLTask(
            lambda: 42,
            dependencies=self._quantum_tasks.copy()
        )

        for ct in [t for t in self._classical_tasks if not t in quantum_dependencies]:
            ct.dependencies.append(gateway_task_quantum)

        self._classical_tasks.append(gateway_task_classical)
        self._quantum_tasks.append(gateway_task_quantum)

        # Start all tasks
        for t in self._classical_tasks:
            t.start()

        for qt in self._quantum_tasks:
            qt.start()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        for t in self.tasks:
            if t.running():
                t.cancel()



from .backends import TabuBackend, SteepestDescentBackend, DwaveBackend, SimulatedAnnealingBackend
from .algorithms import DwaveSolver

__all__ = [
    'DwaveSolver',
    'TabuBackend',
    'DwaveBackend',
    'SimulatedAnnealingBackend',
    'SteepestDescentBackend'
]



import networkx as nx

import numpy as np
import matplotlib.pyplot as plt

from qiskit_algorithms import SamplingMinimumEigensolverResult

from qlauncher.base import Problem


class TSP(Problem):
    """
    Traveling Salesman Problem (TSP) definition.
    """

    def __init__(self, instance: nx.Graph, instance_name: str = "unnamed"):
        """
        Args:
            instance (nx.Graph): Graph representing the TSP instance.
            instance_name (str): Name of the instance.
            quadratic (bool): Whether to use quadratic constraints
        """
        super().__init__(instance=instance, instance_name=instance_name)

    @property
    def setup(self) -> dict:
        return {"instance_name": self.instance_name}

    def _get_path(self) -> str:
        return f"{self.name}@{self.instance_name}"

    def _solution_to_node_chain(self, solution: SamplingMinimumEigensolverResult | str) -> np.ndarray:
        """
        Converts the solution of the TSP problem to a chain of nodes (order of visiting)

        Args:
            solution: Solution of the TSP problem. If type is str, qiskit-style ordering is assumed (MSB first)

        Returns:
            np.ndarray: Solution chain of nodes to visit
        """
        if isinstance(solution, str):
            bitstring = solution[::-1]
        else:
            bitstring = solution.best_measurement["bitstring"][::-1]

        node_count = int(len(bitstring)**0.5)
        chain = []

        for i in range(0, len(bitstring), node_count):
            step_string = bitstring[i: i + node_count]
            chosen_node = np.argmax([int(x) for x in step_string])
            chain.append(chosen_node)

        return np.array(chain)

    def _calculate_solution_cost(self, solution: SamplingMinimumEigensolverResult | str | list[int]) -> float:
        cost = 0
        if isinstance(solution, list):
            chain = solution
        else:
            chain = self._solution_to_node_chain(solution)

        for i in range(len(chain) - 1):
            cost += self.instance[chain[i]][chain[i + 1]]["weight"]
        cost += self.instance[chain[-1]][chain[0]]["weight"]
        return cost

    def _visualize_solution(self, solution: SamplingMinimumEigensolverResult | str | list[int]):
        if isinstance(solution, list):
            chain = solution
        else:
            chain = self._solution_to_node_chain(solution)

        import matplotlib.pyplot as plt

        pos = nx.spring_layout(self.instance, weight=None, seed=42)  # set seed for same node graphs in plt
        problem_edges = list(self.instance.edges)

        marked_edges = []
        for i in range(len(chain) - 1):
            marked_edges.append(set((chain[i], chain[i + 1])))
        marked_edges.append(set((chain[-1], chain[0])))

        draw_colors = []  # Limegreen for marked edges, gray for unmarked
        draw_widths = []  # Draw marked edges thicker

        path_cost = 0
        for edge in problem_edges:
            draw_colors.append("limegreen" if set(edge) in marked_edges else "gray")
            draw_widths.append(2 if set(edge) in marked_edges else 1)

            path_cost += self.instance[edge[0]][edge[1]]["weight"] if set(edge) in marked_edges else 0

        plt.figure(figsize=(8, 6))
        nx.draw(
            self.instance,
            pos,
            edge_color=draw_colors,
            width=draw_widths,
            with_labels=True,
            node_color="skyblue",
            node_size=500,
            font_size=10,
            font_weight="bold",
        )

        labels = nx.get_edge_attributes(self.instance, "weight")
        nx.draw_networkx_edge_labels(
            self.instance,
            pos,
            edge_labels=labels,
            rotate=False,
            font_weight="bold",
            label_pos=0.45,
        )

        plt.title("TSP Solution Visualization")
        plt.suptitle(f"Path cost:{path_cost}")
        plt.show()

    def _visualize_problem(self):
        """
        Show plot of the TSP instance.
        """

        pos = nx.spring_layout(self.instance, weight=None, seed=42)
        plt.figure(figsize=(8, 6))

        nx.draw(
            self.instance,
            pos,
            with_labels=True,
            node_color="skyblue",
            node_size=500,
            edge_color="gray",
            font_size=10,
            font_weight="bold",
        )

        labels = nx.get_edge_attributes(self.instance, "weight")
        nx.draw_networkx_edge_labels(
            self.instance,
            pos,
            edge_labels=labels,
            rotate=False,
            font_weight="bold",
            node_size=500,
            label_pos=0.45,
        )

        plt.title("TSP Instance Visualization")
        plt.show()

    def visualize(self, solution=None):
        if solution is None:
            self._visualize_problem()
        else:
            self._visualize_solution(solution)

    @staticmethod
    def from_preset(instance_name: str = 'default', **kwargs) -> "TSP":
        """
        Generate TSP instance from a preset name.

        Args:
            instance_name (str): Name of the preset instance
            quadratic (bool, optional): Whether to use quadratic constraints. Defaults to False

        Returns:
            TSP: TSP instance
        """
        match instance_name:
            case 'default':
                edge_costs = np.array(
                    [
                        [0, 1, 2, 3],
                        [1, 0, 4, 5],
                        [2, 4, 0, 6],
                        [3, 5, 6, 0]
                    ]
                )
            case _:
                raise ValueError("Unknown instance name")
        G = nx.Graph()
        n = edge_costs.shape[0]
        for i in range(n):
            for j in range(i + 1, n):  # No connections to self
                G.add_edge(i, j, weight=edge_costs[i, j])

        return TSP(instance=G, instance_name=instance_name)

    @staticmethod
    def generate_tsp_instance(num_vertices: int, min_distance: float = 1.0, max_distance: float = 10.0, **kwargs) -> "TSP":
        """
        Generate a random TSP instance.

        Args:
            num_vertices (int): Number of vertices in the graph
            min_distance (float, optional): Minimum distance between vertices. Defaults to 1.0
            max_distance (float, optional): Maximum distance between vertices. Defaults to 10.0
            quadratic (bool, optional): Whether to use quadratic constraints. Defaults to False

        Returns:
            TSP: TSP instance
        """
        if num_vertices < 2:
            raise ValueError("num_vertices must be at least 2")

        if min_distance <= 0:
            raise ValueError("min_distance must be greater than 0")

        g = nx.Graph()
        for i in range(num_vertices):
            for j in range(i + 1, num_vertices):
                g.add_edge(i, j, weight=int(
                    np.random.uniform(min_distance, max_distance)))

        return TSP(instance=g, instance_name="generated")



""" Module for different exceptions used by qlauncher. """


class DependencyError(ImportError):
    """ Error connected with missing optional dependencies and wrong installation. """

    def __init__(self, e: ImportError, install_hint: str = '', *, private: bool = False) -> None:
        if private:
            message = f"""Module "{e.name}" is required but not installed. """ + \
                """The module needs to be installed but it's private. """ + \
                """To get access to module contact the library developers."""
        else:
            message = f"""Module "{e.name}" is required but not installed. Install it with: pip install "qlauncher[{install_hint}]"."""
        super().__init__(message, name=e.name)



from .raw import Raw
from .qatm import QATM
from .jssp import JSSP
from .ec import EC
from .maxcut import MaxCut
from .tsp import TSP
from .graph_coloring import GraphColoring
from .tabular_ml import TabularML
from .molecule import Molecule

__all__ = ['Raw', 'QATM', 'JSSP', 'EC', 'MaxCut', 'TSP', 'GraphColoring',
           'TabularML', 'Molecule']



"""
Utility functions for Hampy objects.
"""
from qiskit.quantum_info import SparsePauliOp, Pauli
from qlauncher.hampy.object import Equation


def shift_affected_qubits(equation: Equation, shift: int) -> Equation:
    """
    Shifts the qubits affected by the equation by the given amount, wrapping around if the index goes out of bounds.

    For each Pauli in the equation hamiltonian, shifts the Pauli string by the given amount.
    i.e (shift = 1) IZIZ -> ZIZI, etc. !Might be unwanted! ZIII -> IIIZ
    Keeps the coefficients the same.

    Args:
        equation: Equation to shift
        shift: Amount to shift by

    Returns:
        Equation: New equation affecting the shifted qubits
    """
    op = equation.hamiltonian

    if shift == 0:
        return equation

    npaulis = []
    ncoeffs = []

    for p_string, coeff in op.label_iter():
        p_string = p_string[shift:] + p_string[:shift]
        npaulis.append(Pauli(data=p_string))
        ncoeffs.append(coeff)

    new_op = SparsePauliOp(npaulis, coeffs=ncoeffs)

    return Equation(new_op)
